# \file
# \ingroup tutorial_roostats
# RooStats Model Inspector
#
# Usage:
# The usage is the same as the StandardXxxDemo.py Python Scripts.
# The macro expects a root file containing a workspace with a ModelConfig and a dataset
#
# ~~~{.py}
# $ ipython3
# %run ModelInspector.py
# #ModelInspector(fileName, workspaceName, modelConfigName, dataSetName)
# ~~~
# Fails with normal python.
# Drag the sliders to adjust the parameters of the model.
# the min and max range of the sliders are used to define the upper & lower variation
# the pointer position of the slider is the central blue curve.
#
# Click the FIT button to
#
# To Do:
#  - Check a better way to exit without crashing.
#  - check boxes to specify which nuisance parameters used in making variation
#  - a button to make the profile inspector plots
#  - a check button to use MINOS errors
#  - have fit button show the covariance matrix from the fit
#  - a button to make the log likelihood plots
#  - a dialog to open the desired file
#  - ability to see the signal and background contributions?
#
# \macro_code
#
#  - Version 1, October 2011
#     - based on tutorial macro by Bertrand Bellenot, Ilka Antcheva
#  - Version 2, November 2011
#     - fixes from Bertrand Bellenot for scrolling window for many parameters
#  - Version 3, April 2024
#     - translation from C++ macros to python3
#
# \author Kyle Cranmer (C++ version), and P. P. (Python translation)

import sys
import ROOT

from enum import Enum


class ETestCommandIdentifiers(Enum):
    HId1 = 1
    HId2 = 2
    HId3 = 3
    HCId1 = 4
    HCId2 = 5
    HSId1 = 6


ETestCmdId = ETestCommandIdentifiers


class ModelInspectorGUI(ROOT.TGMainFrame):
    # private:
    # -fCanvas 	 = TRootEmbeddedCanvas()
    # -fLcan 	 = TGLayoutHints()
    # -fFitFcn 	 = TF1()
    # -fPlot 	 = RooPlot()
    # -fWS 	 = RooWorkspace()
    # -fFile 	 = TFile()
    # -fMC 	 = ModelConfig()
    # -fData 	 = super(RooAbsData)
    # -fFitRes 	 = RooFitResult()
    # -
    # -fSliderList 	 = TList()
    # -fFrameList 	 = TList()
    # -fPlotList = vector("RooPlot *")()
    # -fSliderMap =    ROOT.map("TGTripleHSlider *", " TString " )() # it is an instance
    # -
    # -TSliderMap =    ROOT.map(TGTripleHSlider , char )   # it is a type
    # -fLabelMap =    ROOT.map("TGTripleHSlider*" , "TGLabel *" )()
    # -
    # -fFitButton 	 = TGButton()
    # -fExitButton 	 = TGTextButton()
    # -
    # -# BB: a TGCanvas and a vertical frame are needed for using scrollbars
    # -fCan = 		 TGCanvas()
    # -fVFrame = 		 TGVerticalFrame()
    # -
    # -fHframe0 = fHframe1 = fHframe2 = 		 TGHorizontalFrame()
    # -fBly = fBfly1 = fBfly2 = fBfly3 = 		 TGLayoutHints()
    # -fHslider1 = 		 TGTripleHSlider()
    # -fTbh1 = fTbh2 = fTbh3 = 		 TGTextBuffer()
    # -fCheck1 = fCheck2 = 		 TGCheckButton()

    # public:
    # def __init__(RooWorkspace, ModelConfig, RooAbsData): pass
    # def __del__() : pass

    # def CloseWindow() : pass
    # def DoText( str ): pass
    # def DoSlider(): pass
    # def DoSlider( str ): pass
    # def DoFit(): pass
    # def DoExit(): pass
    # def HandleButtons(): pass

    # ______________________________________________________________________________
    def __del__(self):
        # Clean up

        self.Cleanup()

    # ______________________________________________________________________________
    def CloseWindow(self):
        # Called when window is closed via the window manager.

        del self
        pass

    # ______________________________________________________________________________
    def DoText(self):
        # Handle text entry widgets.

        te = ROOT.BindObject(self.gTQSender, ROOT.TGTextEntry)
        Id = te.WidgetId()
        if Id == ETestCmdId.HId1.value:
            fHslider1.SetPosition(atof(self.fTbh1.GetString()), self.fHslider1.GetMaxPosition())
        elif Id == ETestCmdId.HId2.value:
            fHslider1.SetPointerPosition(atof(self.fTbh2.GetString()))
        elif Id == ETestCmdId.HId3.value:
            fHslider1.SetPosition(self.fHslider1.GetMinPosition(), atof(self.fTbh1.GetString()))

        self.DoSlider()

    # ______________________________________________________________________________
    def DoFit(self):
        self.fFitRes = self.fMC.GetPdf().fitTo(self.fData, Save=True)

        for it in self.fSliderMap:
            param = self.fWS.var(it.second)
            param = self.fFitRes.floatParsFinal().find(it.second.Data())
            it.first.SetPosition(param.getVal() - param.getError(), param.getVal() + param.getError())
            it.first.SetPointerPosition(param.getVal())

        self.DoSlider()

    # ______________________________________________________________________________
    # def DoSlider(text = ""):

    #   print(f".", text)
    #

    # ______________________________________________________________________________
    def DoSlider(self):
        # Handle slider widgets.

        # char buf[32];

        simPdf = ROOT.nullptr
        numCats = 0
        if str(self.fMC.GetPdf().ClassName()) == "RooSimultaneous":
            simPdf = self.fMC.GetPdf()
            channelCat = simPdf.indexCat()
            numCats = channelCat.numTypes()
        else:
            pass

        ###############
        if not simPdf:
            ###############
            # if not SimPdf
            ###############

            # pre loop
            # map<TGTripleHSlider , const char >.iterator it
            it = ROOT.map(TGTripleHSlider, char).iterator  # it # unnecessary

            # try :
            #   del self.fPlot
            # except :
            #   pass
            # self.fPlot = (self.fMC.GetObservables().first()).frame()
            self.fPlot = (self.fMC.GetObservables().first()).frame()
            self.fData.plotOn(self.fPlot)
            normCount = Double_t()

            # high loop
            # it0 = self.fSliderMap.begin()
            for it in self.fSliderMap:
                name = it.second
                self.fWS.var(name).setVal(it.first.GetMaxPosition())
                param = self.fWS.var(name)
                self.fLabelMap[it.first].SetText(
                    ROOT.Form(
                        "{:s} = {:.3}f [{:.3}f,{:.3}f]".format(
                            param.GetName(),
                            it.first.GetPointerPosition(),
                            it.first.GetMinPosition(),
                            it.first.GetMaxPosition(),
                        )
                    )
                )

            normCount = self.fMC.GetPdf().expectedEvents(self.fMC.GetObservables())
            self.fMC.GetPdf().plotOn(
                self.fPlot, ROOT.RooFit.Normalization(normCount, ROOT.RooAbsReal.NumEvent), LineColor="r"
            )

            # low loop
            # it0 = self.fSliderMap.begin()
            for it in self.fSliderMap:
                name = it.second
                self.fWS.var(name).setVal(it.first.GetMinPosition())

            normCount = self.fMC.GetPdf().expectedEvents(self.fMC.GetObservables())
            self.fMC.GetPdf().plotOn(
                self.fPlot, ROOT.RooFit.Normalization(normCount, ROOT.RooAbsReal.NumEvent), LineColor="g"
            )

            # central loop
            # it0 = self.fSliderMap.begin()
            for it in self.fSliderMap:
                name = it.second
                self.fWS.var(name).setVal(it.first.GetPointerPosition())

            normCount = self.fMC.GetPdf().expectedEvents(self.fMC.GetObservables())
            self.fMC.GetPdf().plotOn(
                self.fPlot, ROOT.RooFit.Normalization(normCount, ROOT.RooAbsReal.NumEvent), LineColor="b"
            )
            self.fPlot.Draw()

            self.fCanvas.GetCanvas().Modified()
            self.fCanvas.GetCanvas().Update()
            #########################/
        else:
            #########################/
            # else ( indentation belongs to "if simpdf")
            #########################/
            channelCat = simPdf.indexCat()
            #    TIterator* iter = simPdf->indexCat().typeIterator() ;
            frameIndex = 0
            global gchannelCat
            gchannelCat = channelCat
            for tt in channelCat:
                catName = tt.first

                frameIndex += 1
                self.fCanvas.GetCanvas().cd(frameIndex)

                # pre loop
                pdftmp = simPdf.getPdf(str(catName))
                obstmp = pdftmp.getObservables(self.fMC.GetObservables())
                global ggpdftmp, gobstmp
                ggpdftmp = pdftmp
                gobstmp = obstmp
                # return
                obs = obstmp.first()
                # fplotlist is a template, plotlist is the actual vector<RooPlot> with dim  numCats
                global gframeIndex
                gframeIndex = frameIndex
                global gfPlotList
                gfPlotList = self.fPlotList
                self.fPlot = self.fPlotList[frameIndex - 1]
                if self.fPlot:
                    del self.fPlot
                self.fPlot = obs.frame()

                self.fPlotList[frameIndex - 1] = self.fPlot
                # plotlist[(frameIndex - 1)] = fPlot

                msglevel = ROOT.RooMsgService.instance().globalKillBelow()
                ROOT.RooMsgService.instance().setGlobalKillBelow(ROOT.RooFit.WARNING)
                # """
                self.fData.plotOn(
                    self.fPlot,
                    MarkerSize=1,
                    Cut=ROOT.Form("{:s}=={:s}::{:s}".format(channelCat.GetName(), channelCat.GetName(), str(catName))),
                    DataError="None",
                )
                # """

                # self.fData.plotOn(self.fPlot)

                ROOT.RooMsgService.instance().setGlobalKillBelow(msglevel)

                # normCount = Double_t()

                # high loop
                # it0 = self.fSliderMap.begin()
                for it in self.fSliderMap:
                    name = it.second
                    self.fWS.var(name).setVal(it.first.GetMaxPosition())
                    param = self.fWS.var(name)  # RooRealVar
                    self.fLabelMap[it.first].SetText(
                        ROOT.Form(
                            "{:s} = {:.3f} [{:.3f},{:.3f}]".format(
                                param.GetName(),
                                it.first.GetPointerPosition(),
                                it.first.GetMinPosition(),
                                it.first.GetMaxPosition(),
                            )
                        )
                    )
                # normCount = pdftmp.expectedEvents(obs)
                normCount = pdftmp.expectedEvents(obstmp)
                # normCount = pdftmp.expectedEvents(RooArgSet(obs))
                pdftmp.plotOn(
                    self.fPlot,
                    ROOT.RooFit.Normalization(normCount, ROOT.RooAbsReal.NumEvent),
                    LineColor="r",
                    LineWidth=2,
                )

                # low loop
                # it0 = self.fSliderMap.begin()
                for it in self.fSliderMap:
                    name = it.second
                    self.fWS.var(name).setVal(it.first.GetMinPosition())
                    param = self.fWS.var(name)
                    self.fLabelMap[it.first].SetText(
                        ROOT.Form(
                            "{:s} = {:.3f} [{:.3f},{:.3f}]".format(
                                param.GetName(),
                                it.first.GetPointerPosition(),
                                it.first.GetMinPosition(),
                                it.first.GetMaxPosition(),
                            )
                        )
                    )

                # normCount = pdftmp.expectedEvents(RooArgSet(obs))
                normCount = pdftmp.expectedEvents(obstmp)
                pdftmp.plotOn(
                    self.fPlot,
                    ROOT.RooFit.Normalization(normCount, ROOT.RooAbsReal.NumEvent),
                    LineColor="g",
                    LineWidth=2,
                )

                # central loop
                # it0 = self.fSliderMap.begin()
                for it in self.fSliderMap:
                    name = it.second
                    self.fWS.var(name).setVal(it.first.GetPointerPosition())
                    param = self.fWS.var(name)
                    self.fLabelMap[it.first].SetText(
                        ROOT.Form(
                            "{:s} = {:.3f} [{:.3f},{:.3f}]".format(
                                param.GetName(),
                                it.first.GetPointerPosition(),
                                it.first.GetMinPosition(),
                                it.first.GetMaxPosition(),
                            )
                        )
                    )

                # normCount = pdftmp.expectedEvents(RooArgSet(obs))
                normCount = pdftmp.expectedEvents(obstmp)
                if not self.fFitRes:
                    pass
                    global gnormCount
                    gnormCount = normCount
                    global gpdftmp
                    gpdftmp = pdftmp
                    pdftmp.plotOn(
                        self.fPlot,
                        ROOT.RooFit.Normalization(normCount, ROOT.RooAbsReal.NumEvent),
                        LineColor="b",
                        LineWidth=2,
                    )
                    # pdftmp.plotOn(self.fPlot)
                else:
                    pdftmp.plotOn(
                        self.fPlot,
                        ROOT.RooFit.Normalization(normCount, ROOT.RooAbsReal.NumEvent),
                        ROOT.RooFit.VisualizeError(self.fFitRes, self.fMC.GetNuisanceParameters()),
                        FillColor="y",
                    )
                    pdftmp.plotOn(
                        self.fPlot,
                        ROOT.RooFit.Normalization(normCount, ROOT.RooAbsReal.NumEvent),
                        LineColor="b",
                        LineWidth=2,
                    )
                    msglevel = ROOT.RooMsgService.instance().globalKillBelow()
                    ROOT.RooMsgService.instance().setGlobalKillBelow(ROOT.RooFit.WARNING)
                    self.fData.plotOn(
                        self.fPlot,
                        MarkerSize=1,
                        Cut=ROOT.Form(
                            "{:s}=={:s}::{:s}".format(channelCat.GetName(), channelCat.GetName(), str(catName))
                        ),
                        DataError="None",
                    )

                    ROOT.RooMsgService.instance().setGlobalKillBelow(msglevel)

                self.fPlot.Draw()

            self.fCanvas.GetCanvas().Modified()
            self.fCanvas.GetCanvas().Update()
            ##############/
            # end if(simPdf)
            # return

    # ______________________________________________________________________________
    def HandleButtons(self):

        # Handle different buttons.

        # btn = ROOT.BindObject( ROOT.gTQSender, ROOT.TGTextButton )
        # doesnt' work properly; since we are inside a function a new instance of ROOT is created.
        # then, the ROOT.gTQSender doesn't have any information.
        # self.gTQSender is a stored variable saved after super().__init__(ROOT.gClient.GetRoot(), ...)

        btn = ROOT.BindObject(self.gTQSender, ROOT.TGTextButton)
        Id = btn.WidgetId()
        match (Id):
            case ETestCmdId.HCId1.value:
                self.fHslider1.SetConstrained(self.fCheck1.GetState())
                pass
            case ETestCmdId.HCId2.value:
                self.fHslider1.SetRelative(self.fCheck2.GetState())
                pass
            case _:
                pass

    def DoExit(self):

        print("Exit application...")
        self.gApplication.Terminate(0)
        del self
        sys.exit()

    # ______________________________________________________________________________
    def __init__(self, w, mc, data):
        # ----------------------------------members---------------------
        # --- Principal Members ---
        self.fWS = w
        self.fMC = mc
        self.fData = data
        # --- Principal Members ---
        self.fCanvas = ROOT.TRootEmbeddedCanvas()
        self.fLcan = ROOT.TGLayoutHints()
        self.fFitFcn = ROOT.TF1()
        self.fPlot = ROOT.RooPlot()
        self.fFile = ROOT.TFile()
        self.fFitRes = ROOT.RooFitResult()
        # -
        self.fSliderList = ROOT.TList()
        self.fFrameList = ROOT.TList()
        self.fPlotList = ROOT.std.vector("RooPlot *")()
        self.fSliderMap = ROOT.std.map("TGTripleHSlider *", " TString ")()  # it is an instance
        # -
        # -TSliderMap =    ROOT.map(TGTripleHSlider , char )   # it is a type
        self.fLabelMap = ROOT.std.map("TGTripleHSlider*", "TGLabel *")()
        # -
        self.fFitButton = ROOT.TGButton()
        self.fExitButton = ROOT.TGTextButton()
        # -
        # -# BB: a TGCanvas and a vertical frame are needed for using scrollbars
        self.fCan = ROOT.TGCanvas()
        self.fVFrame = ROOT.TGVerticalFrame()
        # -
        self.fHframe0 = self.fHframe1 = self.fHframe2 = ROOT.TGHorizontalFrame()
        self.fBly = self.fBfly1 = self.fBfly2 = self.fBfly3 = ROOT.TGLayoutHints()
        self.fHslider1 = ROOT.TGTripleHSlider()
        self.fTbh1 = self.fTbh2 = self.fTbh3 = ROOT.TGTextBuffer()
        self.fCheck1 = self.fCheck2 = ROOT.TGCheckButton()
        # -------------------------------------------end of members---------------------
        # debugging
        # global gself
        # gself = self
        #
        # super(TGMainFrame, self).__init__(gClient.GetRoot(), 500, 500)
        # Initialize TGMainFrame and saving its pointers
        super().__init__(ROOT.gClient.GetRoot(), 500, 500)
        self.gTQSender = ROOT.gTQSender
        self.gApplication = ROOT.gApplication

        ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.NumericIntegration)

        simPdf = ROOT.nullptr
        numCats = 1
        # if (strcmp(fMC.GetPdf().ClassName(), "RooSimultaneous") == 0) : non-pythonic syntax
        if self.fMC.GetPdf().ClassName() == "RooSimultaneous":  # simple, pythonic syntax
            print(f"Is a simultaneous PDF")
            simPdf = self.fMC.GetPdf()
            channelCat = simPdf.indexCat()
            print(f" with {channelCat.numTypes()} categories")
            numCats = channelCat.numTypes()
        else:
            print(f"Is not a simultaneous PDF")

        self.fFitRes = ROOT.nullptr
        self.SetCleanup(ROOT.kDeepCleanup)

        # Create an embedded canvas and add it to the main frame with center at x and y
        # and with 30 pixel margin all around
        self.fCanvas = ROOT.TRootEmbeddedCanvas("Canvas", self, 600, 400)
        self.fLcan = ROOT.TGLayoutHints(ROOT.kLHintsExpandX | ROOT.kLHintsExpandY, 10, 10, 10, 10)
        self.AddFrame(self.fCanvas, self.fLcan)
        self.fPlotList.resize(numCats)
        # plotlist = self.fPlotList(numCats) # instead we create an instance of the template
        if numCats > 1:
            self.fCanvas.GetCanvas().Divide(numCats)
            for i in range(numCats):
                self.fCanvas.GetCanvas().cd(i + 1).SetBorderMode(0)
                self.fCanvas.GetCanvas().cd(i + 1).SetGrid()

        # return
        self.fHframe0 = ROOT.TGHorizontalFrame(self, 0, 0, 0)

        self.fCheck1 = ROOT.TGCheckButton(self.fHframe0, "&Constrained", ETestCmdId.HCId1.value)
        self.fCheck2 = ROOT.TGCheckButton(self.fHframe0, "&Relative", ETestCmdId.HCId2.value)
        self.fCheck1.SetState(ROOT.kButtonUp)
        self.fCheck2.SetState(ROOT.kButtonUp)
        self.fCheck1.SetToolTipText("Pointer position constrained to slider sides")
        self.fCheck2.SetToolTipText("Pointer position relative to slider position")

        self.fHframe0.Resize(200, 50)

        self.fHframe2 = ROOT.TGHorizontalFrame(self, 0, 0, 0)

        dp_DoFit = ROOT.TPyDispatcher(self.DoFit)
        self.fFitButton = ROOT.TGTextButton(self.fHframe2, "&Fit")
        self.fFitButton.SetFont("Helvetica")
        self.fFitButton.Connect("Clicked()", "TPyDispatcher", dp_DoFit, "Dispatch()")

        dp_DoExit = ROOT.TPyDispatcher(self.DoExit)
        self.fExitButton = ROOT.TGTextButton(self.fHframe2, "&Exit")
        self.fExitButton.SetFont("Helvetica")
        # self.fExitButton.Connect( "Clicked()", "TPyDispatcher", dp_DoExit , "Dispatch()")
        # doesn't work properly. Break segmentation violation. Full crash.
        self.fExitButton.SetCommand('TPython::Exec( "raise SystemExit" )')

        # dp_CloseWindow = TPyDispatcher( self.CloseWindow)
        # self.Connect("CloseWindow()", "TPyDispatcher", dp_CloseWindow, "Dispatch()")
        self.DontCallClose()

        dp_HandleButtons = ROOT.TPyDispatcher(self.HandleButtons)
        self.fCheck1.Connect("Clicked()", "TPyDispatcher", dp_HandleButtons, "Dispatch()")
        self.fCheck2.Connect("Clicked()", "TPyDispatcher", dp_HandleButtons, "Dispatch()")

        self.fHframe2.Resize(100, 25)

        # --- layout for buttons: top align, equally expand horizontally
        self.fBly = ROOT.TGLayoutHints(ROOT.kLHintsTop | ROOT.kLHintsExpandX, 5, 5, 5, 5)

        # --- layout for the frame: place at bottom, right aligned
        self.fBfly1 = ROOT.TGLayoutHints(ROOT.kLHintsTop | ROOT.kLHintsCenterX, 5, 5, 5, 5)
        self.fBfly2 = ROOT.TGLayoutHints(ROOT.kLHintsTop | ROOT.kLHintsLeft, 5, 5, 5, 5)
        self.fBfly3 = ROOT.TGLayoutHints(ROOT.kLHintsTop | ROOT.kLHintsRight, 5, 5, 5, 5)

        self.fHframe2.AddFrame(self.fFitButton, self.fBfly2)
        self.fHframe2.AddFrame(self.fExitButton, self.fBfly3)

        self.AddFrame(self.fHframe0, self.fBly)
        self.AddFrame(self.fHframe2, self.fBly)

        # Loop over POI & NP, create slider
        # need maps of NP->slider? or just slider->NP
        parameters = ROOT.RooArgSet()
        parameters.add(self.fMC.GetParametersOfInterest())
        parameters.add(self.fMC.GetNuisanceParameters())
        # it = parameters.createIterator() # unnecessary
        param = ROOT.nullptr
        # BB: This is the part needed in order to have scrollbars
        self.fCan = ROOT.TGCanvas(self, 100, 100, ROOT.kFixedSize)
        self.AddFrame(self.fCan, ROOT.TGLayoutHints(ROOT.kLHintsExpandY | ROOT.kLHintsExpandX))
        self.fVFrame = ROOT.TGVerticalFrame(self.fCan.GetViewPort(), 10, 10)
        self.fCan.SetContainer(self.fVFrame)
        # And that's it!
        # Obviously, the parent of other subframes is now fVFrame instead of "self"...

        # while (param := it.Next()): #unnecessary
        for param in parameters:
            print(f"Adding Slider for ", param.GetName())
            hframek = ROOT.TGHorizontalFrame(self.fVFrame, 0, 0, 0)

            hlabel = ROOT.TGLabel(
                hframek, ROOT.Form("{:s} = {:.3f} +{:.3f}".format(param.GetName(), param.getVal(), param.getError()))
            )

            hsliderk = ROOT.TGTripleHSlider(
                hframek,
                0,
                ROOT.kDoubleScaleBoth,
                ETestCmdId.HSId1.value,
                ROOT.kHorizontalFrame,
                ROOT.TGFrame.GetDefaultFrameBackground(),
                False,
                False,
                False,
                False,
            )

            dp_DoSlider = ROOT.TPyDispatcher(self.DoSlider)
            hsliderk.Connect("PointerPositionChanged()", "TPyDispatcher", dp_DoSlider, "Dispatch()")
            hsliderk.Connect("PositionChanged()", "TPyDispatcher", dp_DoSlider, "Dispatch()")
            hsliderk.SetRange(param.getMin(), param.getMax())

            hframek.Resize(200, 25)
            self.fSliderList.Add(hsliderk)
            self.fFrameList.Add(hframek)

            hsliderk.SetPosition(param.getVal() - param.getError(), param.getVal() + param.getError())
            hsliderk.SetPointerPosition(param.getVal())

            hframek.AddFrame(hlabel, self.fBly)  #
            hframek.AddFrame(hsliderk, self.fBly)  #
            self.fVFrame.AddFrame(hframek, self.fBly)
            self.fSliderMap[hsliderk] = param.GetName()
            self.fLabelMap[hsliderk] = hlabel

        # Set main frame name, map sub windows (buttons), initialize layout
        # algorithm via Resize() and map main frame
        self.SetWindowName("RooFit/RooStats Model Inspector")
        self.MapSubwindows()
        self.Resize(self.GetDefaultSize())
        self.MapWindow()

        self.DoSlider()


# ----------------------------------------------------------------------------------------------------
def ModelInspector(infile="", workspaceName="combined", modelConfigName="ModelConfig", dataName="obsData"):

    # -------------------------------------------------------
    # First part is just to access a user-defined file
    # or create the standard example file if it doesn't exist

    filename = ""
    if infile == "":
        filename = "results/example_combined_GaussExample_model.root"
        fileExist = not ROOT.gSystem.AccessPathName(filename)  # note opposite return code
        # if file does not exists generate with histfactory
        if not fileExist:
            # Normally this would be run on the command line
            print(f"will run standard hist2workspace example")
            ROOT.gROOT.ProcessLine(".! prepareHistFactory .")
            ROOT.gROOT.ProcessLine(".! hist2workspace config/example.xml")
            print(f"\n\n---------------------")
            print(f"Done creating example input")
            print(f"---------------------\n\n")

    else:
        filename = infile

    # Bad behaviour of variable, workspace, modelconfig. They get unset after its first call(being whatever)
    # if we declare pointer to the file, workspace, modelconfig, so everything seems to work-out fine.
    Declare = ROOT.gInterpreter.Declare
    Declare(
        """using namespace std;
              using namespace RooFit;
              using namespace RooStats;
              """
    )
    ##################################################
    # Try to open the file:
    # Not to use: file = TFile.Open(filename, "READ" )
    Declare(f'TFile *file = TFile::Open("{filename}");')
    file = ROOT.file

    # if input file was specified but not found, quit
    try:
        file.GetName()
    except ReferenceError:
        print("file wasn't load properly and is a nullptr")
        print(f"\nInput file {filename} was not found")
        return

    # -------------------------------------------------------
    # Tutorial starts here
    # -------------------------------------------------------
    # get the workspace out of the file
    # Not to use :w = file.Get(workspaceName) # RooWorkspace
    ProcessLine = ROOT.gInterpreter.ProcessLine
    ProcessLine(f'RooWorkspace *w = (RooWorkspace *)file->Get("{workspaceName}");')
    w = ROOT.w
    try:
        w.GetName()
    except ReferenceError:
        print(f"Workspace:{workspaceName} wasn't load properly and is a nullptr")
        return

    # get the modelConfig out of the file
    # Not to use: mc = w.obj(modelConfigName) # ModelConfig
    ProcessLine(f'ModelConfig *mc = (ModelConfig *)w->obj("{modelConfigName}");')
    mc = ROOT.mc

    # get the data out of the workspace
    # Not to use: data = w.data(dataName) # RooAbsData
    ProcessLine(f'RooAbsData *DATA = w->data("{dataName}");')
    data = ROOT.DATA

    # make sure ingredients are found
    try:
        mc.GetName()
        print("name", mc.GetPdf())
        mc.GetPdf()
        data.GetName()
    except ReferenceError:
        print(f"ModelConfig:{modelConfigName} wasn't load properly and is a nullptr")
        return
    try:
        data.GetName()
    except ReferenceError:
        print(f"data:{dataName} wasn't load properly and is a nullptr")
        return

    print("running ModelInspectorGUI...")
    ModelInspectorGUI(w, mc, data)


ModelInspector(infile="", workspaceName="combined", modelConfigName="ModelConfig", dataName="obsData")



## \file
## \ingroup tutorial_roofit_main
## \notebook -js
## Basic functionality: demonstrate fitting multiple models using RooMultiPdf and selecting the best one via Discrete
## Profiling method.
##
## \macro_image
## \macro_code
## \macro_output
##
## \date July 2025
## \author Galin Bistrev

import ROOT

x = ROOT.RooRealVar("x", "Observable", 0, 50)


lambda1 = ROOT.RooRealVar("lambda1", "slope1", -0.025, -0.1, -0.02)
expo1 = ROOT.RooExponential("expo1", "Exponential 1", x, lambda1)

c0 = ROOT.RooRealVar("c0", "Cheby coeff 0", -1.0, -1.0, 1.0)
c1 = ROOT.RooRealVar("c1", "Cheby coeff 1", 0.4, 0.05, 0.5)
chebCoeffs = ROOT.RooArgList(c0, c1)
cheb = ROOT.RooChebychev("cheb", "Chebyshev PDF", x, chebCoeffs)

pdfIndex0 = ROOT.RooCategory("pdfIndex0", "pdf index 0")
multiPdf0 = ROOT.RooMultiPdf("multiPdf0", "multiPdf0", pdfIndex0, ROOT.RooArgList(expo1, cheb))


lambdaExtra = ROOT.RooRealVar("lambdaExtra", "extra slope", -0.05, -1.0, -0.01)
expoExtra = ROOT.RooExponential("expoExtra", "extra exponential", x, lambdaExtra)


mean = ROOT.RooRealVar("mean", "shared mean", 25, 0, 50)
sigmaG = ROOT.RooRealVar("sigmaG", "Gaussian width", 2.0, 0.0, 5.0)
sigmaL = ROOT.RooRealVar("sigmaL", "Landau width", 3.0, 1.0, 8.0)

gauss1 = ROOT.RooGaussian("gauss1", "Gaussian", x, mean, sigmaG)
landau1 = ROOT.RooLandau("landau1", "Landau", x, mean, sigmaL)

pdfIndex1 = ROOT.RooCategory("pdfIndex1", "pdf index 1")
multiPdf1 = ROOT.RooMultiPdf("multiPdf1", "multiPdf1", pdfIndex1, ROOT.RooArgList(gauss1, landau1))


sigmaExtra = ROOT.RooRealVar("sigmaExtra", "extra Gaussian width", 3.0, 1.0, 6.0)
gaussExtra = ROOT.RooGaussian("gaussExtra", "extra Gaussian", x, mean, sigmaExtra)


frac0 = ROOT.RooRealVar("frac0", "fraction for cat0", 0.7, 0.0, 1.0)
addPdf0 = ROOT.RooAddPdf("addPdf0", "multiPdf0 + extra expo", ROOT.RooArgList(multiPdf0, gaussExtra), frac0)

frac1 = ROOT.RooRealVar("frac1", "fraction for cat1", 0.5, 0.0, 1.0)
addPdf1 = ROOT.RooAddPdf("addPdf1", "multiPdf1 + extra gauss", ROOT.RooArgList(multiPdf1, expoExtra), frac1)
catIndex = ROOT.RooCategory("catIndex", "Category")
catIndex.defineType("cat0", 0)
catIndex.defineType("cat1", 1)

simPdf = ROOT.RooSimultaneous("simPdf", "simultaneous model", catIndex)
simPdf.addPdf(addPdf0, "cat0")
simPdf.addPdf(addPdf1, "cat1")


data0 = addPdf0.generate(ROOT.RooArgSet(x), 800)
data1 = addPdf1.generate(ROOT.RooArgSet(x), 1000)


frame0 = x.frame()
data0.plotOn(frame0)
addPdf0.plotOn(frame0)
pdfIndex0.setIndex(1)
addPdf0.plotOn(frame0, ROOT.RooFit.LineColor(ROOT.kRed))

frame0.SetTitle("")
frame0.GetXaxis().SetTitle("Observable")
frame0.GetYaxis().SetTitle("Events")

leg0 = ROOT.TLegend(0.6, 0.7, 0.9, 0.9)
leg0.AddEntry(frame0.getObject(0), "Data", "lep")
leg0.AddEntry(frame0.getObject(1), "Expo", "l")
leg0.AddEntry(frame0.getObject(2), "Poly", "l")


frame1 = x.frame()
data1.plotOn(frame1)
addPdf1.plotOn(frame1)
pdfIndex1.setIndex(1)
addPdf1.plotOn(frame1, ROOT.RooFit.LineColor(ROOT.kRed))

frame1.SetTitle("")
frame1.GetXaxis().SetTitle("Observable")
frame1.GetYaxis().SetTitle("Events")

leg1 = ROOT.TLegend(0.6, 0.7, 0.9, 0.9)
leg1.AddEntry(frame1.getObject(0), "Data", "lep")
leg1.AddEntry(frame1.getObject(1), "Gauss", "l")
leg1.AddEntry(frame1.getObject(2), "Landau", "l")


combined_data = ROOT.RooDataSet("data", "combined", ROOT.RooArgSet(x, catIndex))

combined_vars = ROOT.RooArgSet(x, catIndex)

for i in range(data0.numEntries()):
    x.setVal(data0.get(i).getRealValue("x"))
    catIndex.setLabel("cat0")
    combined_data.add(combined_vars)

for i in range(data1.numEntries()):
    x.setVal(data1.get(i).getRealValue("x"))
    catIndex.setLabel("cat1")
    combined_data.add(combined_vars)


nll = simPdf.createNLL(combined_data)
minim = ROOT.RooMinimizer(nll)
minim.setStrategy(1)
minim.setEps(1e-7)
minim.setPrintLevel(-1)


nMeanPoints = 40
meanMin = 17
meanMax = 33

combosToPlot = [[i, j] for i in range(pdfIndex0.numTypes()) for j in range(pdfIndex1.numTypes())]

colors = [ROOT.kRed, ROOT.kBlue, ROOT.kGreen + 2, ROOT.kMagenta, ROOT.kOrange + 7]
markers = [20, 21, 22, 23, 33]

graphs = []
for idx in range(len(combosToPlot)):
    g = ROOT.TGraph(nMeanPoints)
    g.SetLineColor(colors[idx % 5])
    g.SetMarkerColor(colors[idx % 5])
    g.SetMarkerStyle(markers[idx % 5])
    g.SetTitle(f"Combo [{combosToPlot[idx][0]},{combosToPlot[idx][1]}]")
    graphs.append(g)

profileGraph = ROOT.TGraph(nMeanPoints)
profileGraph.SetLineColor(ROOT.kBlack)
profileGraph.SetLineWidth(4)
profileGraph.SetMarkerColor(ROOT.kBlack)
profileGraph.SetMarkerStyle(22)
profileGraph.SetTitle("Profile")
graphs.append(profileGraph)


for i in range(nMeanPoints):
    meanVal = meanMin + i * (meanMax - meanMin) / (nMeanPoints - 1)
    mean.setVal(meanVal)

    for comboIdx, combo in enumerate(combosToPlot):
        pdfIndex0.setIndex(combo[0])
        pdfIndex1.setIndex(combo[1])
        pdfIndex0.setConstant(True)
        pdfIndex1.setConstant(True)
        mean.setConstant(True)
        minim.minimize("Minuit2", "Migrad")
        graphs[comboIdx].SetPoint(i, meanVal, nll.getVal())
        pdfIndex0.setConstant(False)
        pdfIndex1.setConstant(False)
        mean.setConstant(False)

    mean.setConstant(True)
    minim.minimize("Minuit2", "Migrad")
    profileGraph.SetPoint(i, meanVal, nll.getVal())


c = ROOT.TCanvas("c_rf619", "NLL vs Mean for Different Discrete Combinations", 1200, 400)
c.Divide(3, 1)

c.cd(1)
ROOT.gPad.SetLeftMargin(0.15)
frame0.GetYaxis().SetTitleOffset(1.4)
frame0.Draw()
leg0.Draw()

c.cd(2)
ROOT.gPad.SetLeftMargin(0.15)
frame1.GetYaxis().SetTitleOffset(1.4)
frame1.Draw()
leg1.Draw()

c.cd(3)
ROOT.gPad.SetLeftMargin(0.15)
mg = ROOT.TMultiGraph()
for g in graphs:
    mg.Add(g, "PL")
mg.Draw("APL")
mg.GetXaxis().SetTitle("Mean")
mg.GetYaxis().SetTitle("NLL")
ROOT.gPad.BuildLegend()



# \file
# \ingroup tutorial_roostats
# \notebook -js
# SPlot tutorial
#
# This tutorial shows an example of using SPlot to unfold two distributions.
# The physics context for the example is that we want to know
# the isolation distribution for real electrons from Z events
# and fake electrons from QCD.  Isolation is our 'control' variable.
# To unfold them, we need a model for an uncorrelated variable that
# discriminates between Z and QCD.  To do this, we use the invariant
# mass of two electrons.  We model the Z with a Gaussian and the QCD
# with a falling exponential.
#
# Note, since we don't have real data in this tutorial, we need to generate
# toy data.  To do that we need a model for the isolation variable for
# both Z and QCD.  This is only used to generate the toy data, and would
# not be needed if we had real data.
#
# \macro_image
# \macro_code
# \macro_output
#
# \authors Kyle Cranmer (C++ version), and P. P. (Python translation)


import ROOT


def AddModel(wspace):

    # Make models for signal (Higgs) and background (Z+jets and QCD)
    # In real life, this part requires an intelligent modeling
    # of signal and background -- this is only an example.

    # set range of observable
    lowRange, highRange = 0.0, 200.0

    # make a ROOT.RooRealVar for the observables
    invMass = ROOT.RooRealVar("invMass", "M_inv", lowRange, highRange, "GeV")
    isolation = ROOT.RooRealVar("isolation", "isolation", 0.0, 20.0, "GeV")

    # --------------------------------------
    # make 2-d model for Z including the invariant mass
    # distribution  and an isolation distribution which we want to
    # unfold from QCD.
    print(f"make z model")
    # mass model for Z
    mZ = ROOT.RooRealVar("mZ", "Z Mass", 91.2, lowRange, highRange)
    sigmaZ = ROOT.RooRealVar("sigmaZ", "Width of Gaussian", 2, 0, 10, "GeV")
    mZModel = ROOT.RooGaussian("mZModel", "Z+jets Model", invMass, mZ, sigmaZ)
    # we know Z mass
    mZ.setConstant()
    # we leave the width of the Z free during the fit in this example.

    # isolation model for Z.  Only used to generate toy MC.
    # the exponential is of the form exp(c*x).  If we want
    # the isolation to decay an e-fold every R GeV, we use
    # c = -1/R.
    zIsolDecayConst = ROOT.RooConstVar("zIsolDecayConst", "z isolation decay  constant", -1)
    zIsolationModel = ROOT.RooExponential("zIsolationModel", "z isolation model", isolation, zIsolDecayConst)

    # make the combined Z model
    zModel = ROOT.RooProdPdf("zModel", "2-d model for Z", ROOT.RooArgSet(mZModel, zIsolationModel))

    # --------------------------------------
    # make QCD model

    print(f"make qcd model")
    # mass model for QCD.
    # the exponential is of the form exp(c*x).  If we want
    # the mass to decay an e-fold every R GeV, we use
    # c = -1/R.
    # We can leave this parameter free during the fit.
    qcdMassDecayConst = ROOT.RooRealVar(
        "qcdMassDecayConst", "Decay const for QCD mass spectrum", -0.01, -100, 100, "1/GeV"
    )
    qcdMassModel = ROOT.RooExponential("qcdMassModel", "qcd Mass Model", invMass, qcdMassDecayConst)

    # isolation model for QCD.  Only used to generate toy MC
    # the exponential is of the form exp(c*x).  If we want
    # the isolation to decay an e-fold every R GeV, we use
    # c = -1/R.
    qcdIsolDecayConst = ROOT.RooConstVar("qcdIsolDecayConst", "Et resolution constant", -0.1)
    qcdIsolationModel = ROOT.RooExponential("qcdIsolationModel", "QCD isolation model", isolation, qcdIsolDecayConst)

    # make the 2-d model
    qcdModel = ROOT.RooProdPdf("qcdModel", "2-d model for QCD", [qcdMassModel, qcdIsolationModel])

    # combined model
    # These variables represent the number of Z or QCD events
    # They will be fitted.
    zYield = ROOT.RooRealVar("zYield", "fitted yield for Z", 500, 0.0, 5000)
    qcdYield = ROOT.RooRealVar("qcdYield", "fitted yield for QCD", 1000, 0.0, 10000)

    # now make the combined models
    print(f"make full model")
    model = ROOT.RooAddPdf("model", "z+qcd background models", [zModel, qcdModel], [zYield, qcdYield])
    massModel = ROOT.RooAddPdf("massModel", "z+qcd invariant mass model", [mZModel, qcdMassModel], [zYield, qcdYield])

    # interesting for debugging and visualizing the model
    model.graphVizTree("fullModel.dot")

    print(f"import model: ")
    model.Print()

    wspace.Import(model)
    wspace.Import(massModel, RecycleConflictNodes=True)


# Add a toy dataset
def AddData(wspace):

    # get what we need out of the workspace to make toy data
    model = wspace["model"]
    invMass = wspace["invMass"]
    isolation = wspace["isolation"]

    # make the toy data
    print("make data set and import to workspace")
    wspace.Import(model.generate([invMass, isolation]), Rename="data")


# Perform the plot
def DoSPlot(wspace):

    print(f"Calculate sWeights")

    # get what we need out of the workspace to do the fit
    massModel = wspace["massModel"]
    zYield = wspace["zYield"]
    qcdYield = wspace["qcdYield"]
    data = wspace["data"]

    # The sPlot technique requires that we fix the parameters
    # of the model that are not yields after doing the fit.
    #
    # This *could* be done with the lines below, however this is taken care of
    # by the ROOT.RooStats.SPlot constructor (or more precisely the AddSWeight
    # method).
    #
    # sigmaZ = ws["sigmaZ")
    # qcdMassDecayConst = ws["qcdMassDecayConst")
    # sigmaZ.setConstant()
    # qcdMassDecayConst.setConstant()

    ROOT.RooMsgService.instance().setSilentMode(True)

    print(f"\n\n------------------------------------------\nThe dataset before creating sWeights:\n")
    data.Print()

    ROOT.RooMsgService.instance().setGlobalKillBelow(ROOT.RooFit.ERROR)

    # Now we use the SPlot class to add SWeights for the isolation variable to
    # our data set based on fitting the yields to the invariant mass variable.
    # Any keyword arguments will be forwarded to the underlying call to RooAbsPdf::fitTo().
    sData = ROOT.RooStats.SPlot("sData", "An SPlot", data, massModel, [zYield, qcdYield], Strategy=0)

    print(f"\n\nThe dataset after creating sWeights:\n")
    data.Print()

    # Check that our weights have the desired properties

    print("\n\n------------------------------------------\n\nCheck SWeights:")
    print("Yield of Z is\t", zYield.getVal(), ".  From sWeights it is ")
    print(sData.GetYieldFromSWeight("zYield"))

    print("Yield of QCD is\t", qcdYield.getVal(), ".  From sWeights it is : ")
    print(sData.GetYieldFromSWeight("qcdYield"))

    for i in range(10):
        print("Weight for event: ", i, sData.GetSWeight(i, "zYield"))
        print("qcd Weight: ", sData.GetSWeight(i, "qcdYield"))
        print("Total Weight: ", sData.GetSumOfEventSWeight(i))

    # import this new dataset with sWeights
    wspace.Import(data, Rename="dataWithSWeights")

    ROOT.RooMsgService.instance().setGlobalKillBelow(ROOT.RooFit.INFO)


def MakePlots(wspace):

    # Here we make plots of the discriminating variable (invMass) after the fit
    # and of the control variable (isolation) after unfolding with sPlot.
    # make our canvas
    cdata = ROOT.TCanvas("sPlot", "sPlot demo", 400, 600)
    cdata.Divide(1, 3)

    # get what we need out of the workspace
    model = wspace["model"]
    zModel = wspace["zModel"]
    qcdModel = wspace["qcdModel"]

    isolation = wspace["isolation"]
    invMass = wspace["invMass"]

    # note, we get the dataset with sWeights
    data = wspace["dataWithSWeights"]

    # create weighted data sets
    dataw_qcd = ROOT.RooDataSet(data.GetName(), data.GetTitle(), data.get(), Import=data, WeightVar="qcdYield_sw")
    dataw_z = ROOT.RooDataSet(data.GetName(), data.GetTitle(), data.get(), Import=data, WeightVar="zYield_sw")

    # plot invMass for data with full model and individual components overlaid
    # cdata = TCanvas()
    cdata.cd(1)
    frame = invMass.frame(Title="Fit of model to discriminating variable")
    data.plotOn(frame)
    model.plotOn(frame, Name="FullModel")
    model.plotOn(frame, Components=zModel, LineStyle="--", LineColor="r", Name="ZModel")
    model.plotOn(frame, Components=qcdModel, LineStyle="--", LineColor="g", Name="QCDModel")

    leg = ROOT.TLegend(0.11, 0.5, 0.5, 0.8)
    leg.AddEntry(frame.findObject("FullModel"), "Full model", "L")
    leg.AddEntry(frame.findObject("ZModel"), "Z model", "L")
    leg.AddEntry(frame.findObject("QCDModel"), "QCD model", "L")
    leg.SetBorderSize(0)
    leg.SetFillStyle(0)

    frame.Draw()
    leg.DrawClone()

    # Now use the sWeights to show isolation distribution for Z and QCD.
    # The SPlot class can make this easier, but here we demonstrate in more
    # detail how the sWeights are used.  The SPlot class should make this
    # very easy and needs some more development.

    # Plot isolation for Z component.
    # Do this by plotting all events weighted by the sWeight for the Z component.
    # The SPlot class adds a new variable that has the name of the corresponding
    # yield + "_sw".
    cdata.cd(2)

    frame2 = isolation.frame(Title="Isolation distribution with s weights to project out Z")
    # Since the data are weighted, we use SumW2 to compute the errors.
    dataw_z.plotOn(frame2, DataError="SumW2")
    zModel.plotOn(frame2, LineStyle="--", LineColor="r")

    frame2.Draw()

    # Plot isolation for QCD component.
    # Eg. plot all events weighted by the sWeight for the QCD component.
    # The SPlot class adds a new variable that has the name of the corresponding
    # yield + "_sw".
    cdata.cd(3)
    frame3 = isolation.frame(Title="Isolation distribution with s weights to project out QCD")
    dataw_qcd.plotOn(frame3, DataError="SumW2")
    qcdModel.plotOn(frame3, LineStyle="--", LineColor="g")

    frame3.Draw()

    cdata.SaveAs("rs301_splot.png")


def rs301_splot():

    # Create a workspace to manage the project.
    wspace = ROOT.RooWorkspace("myWS")

    # add the signal and background models to the workspace.
    # Inside this function you will find a description of our model.
    AddModel(wspace)

    # add some toy data to the workspace
    AddData(wspace)

    # do sPlot.
    # This will make a new dataset with sWeights added for every event.
    DoSPlot(wspace)

    # Make some plots showing the discriminating variable and
    # the control variable after unfolding.
    MakePlots(wspace)


rs301_splot()



## \file
## \ingroup tutorial_dataframe
## \notebook -nodraw
## Use Range to limit the amount of data processed.
##
## This tutorial shows how to express the concept of ranges when working with the RDataFrame.
##
## \macro_code
## \macro_output
##
## \date March 2017
## \author Danilo Piparo (CERN)

import ROOT

def fill_tree(treeName, fileName):
    df = ROOT.RDataFrame(100)
    df.Define("b1", "(int) rdfentry_")\
      .Define("b2", "(float) rdfentry_ * rdfentry_").Snapshot(treeName, fileName)


# We prepare an input tree to run on
fileName = "df006_ranges_py.root"
treeName = "myTree"

fill_tree(treeName, fileName)

# We read the tree from the file and create a RDataFrame.
d = ROOT.RDataFrame(treeName, fileName)

# ## Usage of ranges
# Now we'll count some entries using ranges
c_all = d.Count()

# This is how you can express a range of the first 30 entries
d_0_30 = d.Range(30) 
c_0_30 = d_0_30.Count()

# This is how you pick all entries from 15 onwards
d_15_end = d.Range(15, 0)
c_15_end = d_15_end.Count()

# We can use a stride too, in this case we pick an event every 3 entries
d_15_end_3 = d.Range(15, 0, 3)
c_15_end_3 = d_15_end_3.Count()

# The Range here acts first on the (whole) RDataFrame graph:
# Not only actions (like Count) but also filters and new columns can be added to it.
d_0_50 = d.Range(50)
c_0_50_odd_b1 = d_0_50.Filter("1 == b1 % 2").Count()

# An important thing to notice is that the counts of a filter are relative to the
# number of entries a filter "sees". Therefore, if a Range depends on a filter,
# the Range will act on the entries passing the filter only.
c_0_3_after_even_b1 = d.Filter("0 == b1 % 2").Range(0, 3).Count()

# Ok, time to wrap up: let's print all counts!
print("Usage of ranges:")
print(" - All entries:", c_all.GetValue())
print(" - Entries from 0 to 30:", c_0_30.GetValue())
print(" - Entries from 15 onwards:", c_15_end.GetValue())
print(" - Entries from 15 onwards in steps of 3:", c_15_end_3.GetValue())
print(" - Entries from 0 to 50, odd only:", c_0_50_odd_b1.GetValue())
print(" - First three entries of all even entries:", c_0_3_after_even_b1.GetValue())



## \file
## \ingroup tutorial_math
## \notebook
## Example of first few Legendre Polynomials. Inspired by work of Lorenzo Moneta.
##
## \macro_image
## \macro_code
##
## \author Alberto Ferro, Massimiliano Galli


import ROOT

ROOT.gSystem.Load("libMathMore")
Canvas = ROOT.TCanvas("DistCanvas", "Legendre polynomials example", 10, 10, 750, 600)
Canvas.SetGrid()
leg = ROOT.TLegend(0.5, 0.7, 0.4, 0.89)

L = []
for nu in range(5):
    f = ROOT.TF1("L_0", "ROOT::Math::legendre([0],x)", -1, 1)
    f.SetParameters(nu, 0.0)
    f.SetLineStyle(ROOT.kSolid)
    f.SetLineWidth(2)
    f.SetLineColor(nu+1)
    L.append(f)

L[0].SetMaximum(1)
L[0].SetMinimum(-1)
L[0].SetTitle("Legendre polynomials")

for idx, val in enumerate(L):
    leg.AddEntry(val, " L_{}(x)".format(idx), "l")
    if idx == 0:
        val.Draw()
    else:
        val.Draw("same")

leg.Draw("same")





## \file
## \ingroup tutorial_roofit_main
## \notebook
## Extended maximum likelihood fit in multiple ranges.
##
## \macro_code
## \macro_output
##
## \date March 2021
## \authors Harshal Shende, Stephan Hageboeck (C++ version)

import ROOT


# Setup component pdfs
# ---------------------

# Declare observable x
x = ROOT.RooRealVar("x", "x", 0, 11)

# Create two Gaussian PDFs g1(x,mean1,sigma) anf g2(x,mean2,sigma) and their parameters
mean = ROOT.RooRealVar("mean", "mean of gaussians", 5)
sigma1 = ROOT.RooRealVar("sigma1", "width of gaussians", 0.5)
sigma2 = ROOT.RooRealVar("sigma2", "width of gaussians", 1)

sig1 = ROOT.RooGaussian("sig1", "Signal component 1", x, mean, sigma1)
sig2 = ROOT.RooGaussian("sig2", "Signal component 2", x, mean, sigma2)

# Build Chebychev polynomial pdf
a0 = ROOT.RooRealVar("a0", "a0", 0.5, 0.0, 1.0)
a1 = ROOT.RooRealVar("a1", "a1", 0.2, 0.0, 1.0)
bkg = ROOT.RooChebychev("bkg", "Background", x, [a0, a1])

# Sum the signal components into a composite signal pdf
sig1frac = ROOT.RooRealVar("sig1frac", "fraction of component 1 in signal", 0.8, 0.0, 1.0)
sig = ROOT.RooAddPdf("sig", "Signal", [sig1, sig2], sig1frac)


# Extend the pdfs
# -----------------------------

# Define signal range in which events counts are to be defined
x.setRange("signalRange", 4, 6)

# Associated nsig/nbkg as expected number of events with sig/bkg _in_the_range_ "signalRange"
nsig = ROOT.RooRealVar("nsig", "number of signal events in signalRange", 500, 0.0, 10000)
nbkg = ROOT.RooRealVar("nbkg", "number of background events in signalRange", 500, 0, 10000)

# Use AddPdf to extend the model. Giving as many coefficients as pdfs switches on extension.
model = ROOT.RooAddPdf("model", "(g1+g2)+a", [bkg, sig], [nbkg, nsig])

# Sample data, fit model
# -------------------------------------------

# Generate 1000 events from model so that nsig,nbkg come out to numbers <<500 in fit
data = model.generate(x, 1000)

canv = ROOT.TCanvas("Canvas", "Canvas", 1500, 600)
canv.Divide(3, 1)

# Fit full range
#  -------------------------------------------

# Perform unbinned ML fit to data, full range

# IMPORTANT:
# The model needs to be copied when fitting with different ranges because
# the interpretation of the coefficients is tied to the fit range
# that's used in the first fit
canv.cd(1)

model1 = ROOT.RooAddPdf(model)
r = model1.fitTo(data, Save=True, PrintLevel=-1)
r.Print()

frame = x.frame(Title="Full range fitted")
data.plotOn(frame)
model1.plotOn(frame, VisualizeError=r)
model1.plotOn(frame)
model1.paramOn(frame)
frame.Draw()


# Fit in two regions
# -------------------------------------------

canv.cd(2)
x.setRange("left", 0.0, 4.0)
x.setRange("right", 6.0, 10.0)

model2 = ROOT.RooAddPdf(model)
r2 = model2.fitTo(data, Range="left,right", Save=True, PrintLevel=-1)
r2.Print()

frame2 = x.frame(Title="Fit in left/right sideband")
data.plotOn(frame2)
model2.plotOn(frame2, VisualizeError=r2)
model2.plotOn(frame2)
model2.paramOn(frame2)
frame2.Draw()


# Fit in one region
# -------------------------------------------
# Note how restricting the region to only the left tail increases
# the fit uncertainty

canv.cd(3)
x.setRange("leftToMiddle", 0.0, 5.0)

model3 = ROOT.RooAddPdf(model)
r3 = model3.fitTo(data, Range="leftToMiddle", Save=True, PrintLevel=-1)
r3.Print()

frame3 = x.frame(Title="Fit from left to middle")
data.plotOn(frame3)
model3.plotOn(frame3, VisualizeError=r3)
model3.plotOn(frame3)
model3.paramOn(frame3)
frame3.Draw()

canv.Draw()

canv.SaveAs("rf204a_extendedLikelihood.png")



## \file
## \ingroup tutorial_roofit_main
## \notebook
## Fill RooDataSet/RooDataHist in RDataFrame.
##
## This tutorial shows how to fill RooFit data classes directly from RDataFrame.
## Using two small helpers, we tell RDataFrame where the data has to go.
##
## \macro_code
## \macro_output
##
## \date July 2021
## \author Harshal Shende, Stephan Hageboeck (C++ version)

import ROOT
import math


# Set up
# ------------------------

# We create an RDataFrame with two columns filled with 2 million random numbers.
df = ROOT.RDataFrame(2000000).Define("x", "gRandom->Uniform(-5.,  5.)").Define("y", "gRandom->Gaus(1., 3.)")


# We create RooFit variables that will represent the dataset.
x = ROOT.RooRealVar("x", "x", -5.0, 5.0)
y = ROOT.RooRealVar("y", "y", -50.0, 50.0)
x.setBins(10)
y.setBins(20)


# Booking the creation of RooDataSet / RooDataHist in RDataFrame
# ----------------------------------------------------------------

# Method 1:
# ---------
# We directly book the RooDataSetHelper action.
# We need to pass
# - the RDataFrame column types as template parameters
# - the constructor arguments for RooDataSet (they follow the same syntax as the usual RooDataSet constructors)
# - the column names that RDataFrame should fill into the dataset

# NOTE: RDataFrame columns are matched to RooFit variables by position, *not by name*!
#
# The returned object is not yet a RooDataSet, but an RResultPtr that will be
# lazy-evaluated once you call GetValue() on it. We will only evaluate the
# RResultPtr once all other RDataFrame related actions are declared. This way
# we trigger the event loop computation only once, which will improve the
# runtime significantly.
#
# To learn more about lazy actions, see:
#     https://root.cern/doc/master/classROOT_1_1RDataFrame.html#actions
roo_data_set_result = df.Book(
    ROOT.std.move(ROOT.RooDataSetHelper("dataset", "Title of dataset", ROOT.RooArgSet(x, y))), ("x", "y")
)

# Method 2:
# ---------
# We first declare the RooDataHistHelper
rdhMaker = ROOT.RooDataHistHelper("dataset", "Title of dataset", ROOT.RooArgSet(x, y))

# Then, we move it into an RDataFrame action:
roo_data_hist_result = df.Book(ROOT.std.move(rdhMaker), ("x", "y"))


# Run it and inspect the results
# -------------------------------

# At this point, all RDF actions were defined (namely, the `Book` operations),
# so we can get values from the RResultPtr objects, triggering the event loop
# and getting the actual RooFit data objects.
roo_data_set = roo_data_set_result.GetValue()
roo_data_hist = roo_data_hist_result.GetValue()

# Let's inspect the dataset / datahist.


def print_data(data):
    print("")
    data.Print()
    for i in range(min(data.numEntries(), 20)):
        print(
            "("
            + ", ".join(["{0:8.3f}".format(var.getVal()) for var in data.get(i)])
            + ", )  weight={0:10.3f}".format(data.weight())
        )

    print("mean(x) = {0:.3f}".format(data.mean(x)) + "\tsigma(x) = {0:.3f}".format(math.sqrt(data.moment(x, 2.0))))
    print("mean(y) = {0:.3f}".format(data.mean(y)) + "\tsigma(y) = {0:.3f}\n".format(math.sqrt(data.moment(y, 2.0))))


print_data(roo_data_set)
print_data(roo_data_hist)



## \file
## \ingroup tutorial_roofit_main
## \notebook -nodraw
## Numeric algorithm tuning: configuration and customization of how MC sampling algorithms
## on specific pdfs are executed
##
## \macro_code
## \macro_output
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C++ version)

import ROOT


# Adjust global MC sampling strategy
# ------------------------------------------------------------------

# Example pdf for use below
x = ROOT.RooRealVar("x", "x", 0, 10)
model = ROOT.RooChebychev("model", "model", x, [0.0, 0.5, -0.1])

# Change global strategy for 1D sampling problems without conditional observable
# (1st kFALSE) and without discrete observable (2nd kFALSE) from ROOT.RooFoamGenerator,
# ( an interface to the ROOT.TFoam MC generator with adaptive subdivisioning strategy ) to ROOT.RooAcceptReject,
# a plain accept/reject sampling algorithm [ ROOT.RooFit default before
# ROOT 5.23/04 ]
ROOT.RooAbsPdf.defaultGeneratorConfig().method1D(False, False).setLabel("RooAcceptReject")

# Generate 10Kevt using ROOT.RooAcceptReject
data_ar = model.generate({x}, 10000, Verbose=True)
data_ar.Print()

# Adjusting default config for a specific pdf
# -------------------------------------------------------------------------------------

# Another possibility: associate custom MC sampling configuration as default for object 'model'
# The kTRUE argument will install a clone of the default configuration as specialized configuration
# for self model if none existed so far
model.specialGeneratorConfig(True).method1D(False, False).setLabel("RooFoamGenerator")

# Adjusting parameters of a specific technique
# ---------------------------------------------------------------------------------------

# Adjust maximum number of steps of ROOT.RooIntegrator1D in the global
# default configuration
ROOT.RooAbsPdf.defaultGeneratorConfig().getConfigSection("RooAcceptReject").setRealValue("nTrial1D", 2000)

# Example of how to change the parameters of a numeric integrator
# (Each config section is a ROOT.RooArgSet with ROOT.RooRealVars holding real-valued parameters
#  and ROOT.RooCategories holding parameters with a finite set of options)
model.specialGeneratorConfig().getConfigSection("RooFoamGenerator").setRealValue("chatLevel", 1)

# Generate 10Kevt using ROOT.RooFoamGenerator (FOAM verbosity increased
# with above chatLevel adjustment for illustration purposes)
data_foam = model.generate({x}, 10000, Verbose=True)
data_foam.Print()



## \file
## \ingroup tutorial_roofit_main
## \notebook
## Organization and simultaneous fits: using simultaneous pdfs to describe simultaneous
## fits to multiple datasets
##
## \macro_image
## \macro_code
## \macro_output
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C++ version)

import ROOT

# Create model for physics sample
# -------------------------------------------------------------

# Create observables
x = ROOT.RooRealVar("x", "x", -8, 8)

# Construct signal pdf
mean = ROOT.RooRealVar("mean", "mean", 0, -8, 8)
sigma = ROOT.RooRealVar("sigma", "sigma", 0.3, 0.1, 10)
gx = ROOT.RooGaussian("gx", "gx", x, mean, sigma)

# Construct background pdf
a0 = ROOT.RooRealVar("a0", "a0", -0.1, -1, 1)
a1 = ROOT.RooRealVar("a1", "a1", 0.004, -1, 1)
px = ROOT.RooChebychev("px", "px", x, [a0, a1])

# Construct composite pdf
f = ROOT.RooRealVar("f", "f", 0.2, 0.0, 1.0)
model = ROOT.RooAddPdf("model", "model", [gx, px], [f])

# Create model for control sample
# --------------------------------------------------------------

# Construct signal pdf.
# NOTE that sigma is shared with the signal sample model
mean_ctl = ROOT.RooRealVar("mean_ctl", "mean_ctl", -3, -8, 8)
gx_ctl = ROOT.RooGaussian("gx_ctl", "gx_ctl", x, mean_ctl, sigma)

# Construct the background pdf
a0_ctl = ROOT.RooRealVar("a0_ctl", "a0_ctl", -0.1, -1, 1)
a1_ctl = ROOT.RooRealVar("a1_ctl", "a1_ctl", 0.5, -0.1, 1)
px_ctl = ROOT.RooChebychev("px_ctl", "px_ctl", x, [a0_ctl, a1_ctl])

# Construct the composite model
f_ctl = ROOT.RooRealVar("f_ctl", "f_ctl", 0.5, 0.0, 1.0)
model_ctl = ROOT.RooAddPdf("model_ctl", "model_ctl", [gx_ctl, px_ctl], [f_ctl])

# Generate events for both samples
# ---------------------------------------------------------------

# Generate 1000 events in x and y from model
data = model.generate({x}, 1000)
data_ctl = model_ctl.generate({x}, 2000)

# Create index category and join samples
# ---------------------------------------------------------------------------

# Define category to distinguish physics and control samples events
sample = ROOT.RooCategory("sample", "sample")
sample.defineType("physics")
sample.defineType("control")

# Construct combined dataset in (x,sample)
combData = ROOT.RooDataSet(
    "combData",
    "combined data",
    {x},
    Index=sample,
    Import={"physics": data, "control": data_ctl},
)

# Construct a simultaneous pdf in (x, sample)
# -----------------------------------------------------------------------------------

# Construct a simultaneous pdf using category sample as index: associate model
# with the physics state and model_ctl with the control state
simPdf = ROOT.RooSimultaneous("simPdf", "simultaneous pdf", {"physics": model, "control": model_ctl}, sample)

# Perform a simultaneous fit
# ---------------------------------------------------

# Perform simultaneous fit of model to data and model_ctl to data_ctl
fitResult = simPdf.fitTo(combData, PrintLevel=-1, Save=True)
fitResult.Print()

# Plot model slices on data slices
# ----------------------------------------------------------------

# Make a frame for the physics sample
frame1 = x.frame(Title="Physics sample")

# Plot all data tagged as physics sample
slicedData1 = combData.reduce(Cut="sample==sample::physics")
slicedData1.plotOn(frame1)

# Plot "physics" slice of simultaneous pdf.
simPdf.getPdf("physics").plotOn(frame1)
simPdf.getPdf("physics").plotOn(frame1, Components="px", LineStyle="--")

# The same plot for the control sample slice. We do this with a different
# approach, using the data slice for the projection. This approach is more
# general, because you can plot sums of slices by using logical or in the
# Cut() command.
# NBL You _must_ project the sample index category with data using ProjWData
# as a RooSimultaneous makes no prediction on the shape in the index category
# and can thus not be integrated.
# In other words: Since the PDF doesn't know the number of events in the different
# category states, it doesn't know how much of each component it has to project out.
# This information is read from the data.
frame2 = x.frame(Title="Control sample")
slicedData2 = combData.reduce(Cut="sample==sample::control")
slicedData2.plotOn(frame2)
simPdf.plotOn(frame2, ProjWData=(sample, slicedData2))
simPdf.plotOn(frame2, Components="px_ctl", ProjWData=(sample, slicedData2), LineStyle="--")

# The same plot for all the phase space. Here, we can just use the original
# combined dataset.
frame3 = x.frame(Title="Both samples")
combData.plotOn(frame3)
simPdf.plotOn(frame3, ProjWData=(sample, combData))
simPdf.plotOn(frame3, Components="px,px_ctl", ProjWData=(sample, combData), LineStyle="--")

c = ROOT.TCanvas("rf501_simultaneouspdf", "rf501_simultaneouspdf", 1200, 400)
c.Divide(3)


def draw(i, frame):
    c.cd(i)
    ROOT.gPad.SetLeftMargin(0.15)
    frame.GetYaxis().SetTitleOffset(1.4)
    frame.Draw()


draw(1, frame1)
draw(2, frame2)
draw(3, frame3)

c.SaveAs("rf501_simultaneouspdf.png")



## \file
## \ingroup tutorial_roofit_main
## \notebook -js
## Implementing the Barlow-Beeston method for taking into account the statistical
## uncertainty of a Monte-Carlo fit template.
##
## \macro_image
## \macro_code
## \macro_output
##
## Based on a demo by Wouter Verkerke
## \date June 2021
## \author Harshal Shende, Stephan Hageboeck (C++ version)


import ROOT

# First, construct a likelihood model with a Gaussian signal on top of a uniform background
x = ROOT.RooRealVar("x", "x", -20, 20)
x.setBins(25)

meanG = ROOT.RooRealVar("meanG", "meanG", 1, -10, 10)
sigG = ROOT.RooRealVar("sigG", "sigG", 1.5, -10, 10)
g = ROOT.RooGaussian("g", "Gauss", x, meanG, sigG)
u = ROOT.RooUniform("u", "Uniform", x)


# Generate the data to be fitted
sigData = g.generate(x, 50)
bkgData = u.generate(x, 1000)

sumData = ROOT.RooDataSet("sumData", "Gauss + Uniform", x)
sumData.append(sigData)
sumData.append(bkgData)


# Make histogram templates for signal and background.
# Let's take a signal distribution with low statistics and a more accurate
# background distribution.
# Normally, these come from Monte Carlo simulations, but we will just generate them.
dh_sig = g.generateBinned(x, 50)
dh_bkg = u.generateBinned(x, 10000)


#  Case 0 - 'Rigid templates'

# Construct histogram shapes for signal and background
p_h_sig = ROOT.RooHistFunc("p_h_sig", "p_h_sig", x, dh_sig)
p_h_bkg = ROOT.RooHistFunc("p_h_bkg", "p_h_bkg", x, dh_bkg)

# Construct scale factors for adding the two distributions
Asig0 = ROOT.RooRealVar("Asig", "Asig", 1, 0.01, 5000)
Abkg0 = ROOT.RooRealVar("Abkg", "Abkg", 1, 0.01, 5000)

# Construct the sum model
model0 = ROOT.RooRealSumPdf("model0", "model0", [p_h_sig, p_h_bkg], [Asig0, Abkg0], True)


#  Case 1 - 'Barlow Beeston'

# Construct parameterized histogram shapes for signal and background
p_ph_sig1 = ROOT.RooParamHistFunc("p_ph_sig", "p_ph_sig", dh_sig, x)
p_ph_bkg1 = ROOT.RooParamHistFunc("p_ph_bkg", "p_ph_bkg", dh_bkg, x)

Asig1 = ROOT.RooRealVar("Asig", "Asig", 1, 0.01, 5000)
Abkg1 = ROOT.RooRealVar("Abkg", "Abkg", 1, 0.01, 5000)

# Construct the sum of these
model_tmp = ROOT.RooRealSumPdf("sp_ph", "sp_ph", [p_ph_sig1, p_ph_bkg1], [Asig1, Abkg1], True)

# Construct the subsidiary poisson measurements constraining the histogram parameters
# These ensure that the bin contents of the histograms are only allowed to vary within
# the statistical uncertainty of the Monte Carlo.
hc_sig = ROOT.RooHistConstraint("hc_sig", "hc_sig", p_ph_sig1)
hc_bkg = ROOT.RooHistConstraint("hc_bkg", "hc_bkg", p_ph_bkg1)

# Construct the joint model with template PDFs and constraints
model1 = ROOT.RooProdPdf("model1", "model1", {hc_sig, hc_bkg}, Conditional=(model_tmp, x))


#  Case 2 - 'Barlow Beeston' light (one parameter per bin for all samples)

# Construct the histogram shapes, using the same parameters for signal and background
# This requires passing the first histogram to the second, so that their common parameters
# can be re-used.
# The first ParamHistFunc will create one parameter per bin, such as `p_ph_sig2_gamma_bin_0`.
# This allows bin 0 to fluctuate up and down.
# Then, the SAME parameters are connected to the background histogram, so the bins fluctuate
# synchronously. This reduces the number of parameters.
p_ph_sig2 = ROOT.RooParamHistFunc("p_ph_sig2", "p_ph_sig2", dh_sig, x)
p_ph_bkg2 = ROOT.RooParamHistFunc("p_ph_bkg2", "p_ph_bkg2", dh_bkg, x, p_ph_sig2, True)

Asig2 = ROOT.RooRealVar("Asig", "Asig", 1, 0.01, 5000)
Abkg2 = ROOT.RooRealVar("Abkg", "Abkg", 1, 0.01, 5000)

# As before, construct the sum of signal2 and background2
model2_tmp = ROOT.RooRealSumPdf("sp_ph", "sp_ph", [p_ph_sig2, p_ph_bkg2], [Asig2, Abkg2], True)

# Construct the subsidiary poisson measurements constraining the statistical fluctuations
hc_sigbkg = ROOT.RooHistConstraint("hc_sigbkg", "hc_sigbkg", {p_ph_sig2, p_ph_bkg2})

# Construct the joint model
model2 = ROOT.RooProdPdf("model2", "model2", hc_sigbkg, Conditional=(model2_tmp, x))


# ************ Fit all models to data and plot *********************

result0 = model0.fitTo(sumData, PrintLevel=0, Save=True)
result1 = model1.fitTo(sumData, PrintLevel=0, Save=True)
result2 = model2.fitTo(sumData, PrintLevel=0, Save=True)


can = ROOT.TCanvas("can", "", 1500, 600)
can.Divide(3, 1)

pt = ROOT.TPaveText(-19.5, 1, -2, 25)
pt.SetFillStyle(0)
pt.SetBorderSize(0)


can.cd(1)
frame = x.frame(Title="No template uncertainties")
# Plot data to enable automatic determination of model0 normalisation:
sumData.plotOn(frame)
model0.plotOn(frame, LineColor="b", VisualizeError=result0)
# Plot data again to show it on top of model0 error bands:
sumData.plotOn(frame)
# Plot model components
model0.plotOn(frame, LineColor="b")
p_ph_sig_set = {p_h_sig}
p_ph_bkg_set = {p_h_bkg}
model0.plotOn(frame, Components=p_ph_sig_set, LineColor="kAzure")
model0.plotOn(frame, Components=p_ph_bkg_set, LineColor="r")
model0.paramOn(frame)

sigData.plotOn(frame, MarkerColor="b")
frame.Draw()

pt_text1 = [
    "No template uncertainties",
    "are taken into account.",
    "This leads to low errors",
    "for the parameters A, since",
    "the only source of errors",
    "are the data statistics.",
]
for text in pt_text1:
    pt.AddText(text)

pt.DrawClone()


can.cd(2)
frame = x.frame(Title="Barlow Beeston for Sig & Bkg separately")
sumData.plotOn(frame)
model1.plotOn(frame, LineColor="b", VisualizeError=result1)
# Plot data again to show it on top of error bands:
sumData.plotOn(frame)
model1.plotOn(frame, LineColor="b")
p_ph_sig1_set = {p_ph_sig1}
p_ph_bkg1_set = {p_ph_bkg1}
model1.plotOn(frame, Components=p_ph_sig1_set, LineColor="kAzure")
model1.plotOn(frame, Components=p_ph_bkg1_set, LineColor="r")
model1.paramOn(frame, Parameters={Asig1, Abkg1})

sigData.plotOn(frame, MarkerColor="b")
frame.Draw()

pt.Clear()
pt_text2 = [
    "With gamma parameters, the",
    "signal & background templates",
    "can adapt to the data.",
    "Note how the blue signal",
    "template changes its shape.",
    "This leads to higher errors",
    "of the scale parameters A.",
]

for text in pt_text2:
    pt.AddText(text)

pt.DrawClone()

can.cd(3)
frame = x.frame(Title="Barlow Beeston light for (Sig+Bkg)")
sumData.plotOn(frame)
model2.plotOn(frame, LineColor="b", VisualizeError=result2)
# Plot data again to show it on top of model0 error bands:
sumData.plotOn(frame)
model2.plotOn(frame, LineColor="b")
p_ph_sig2_set = {p_ph_sig2}
p_ph_bkg2_set = {p_ph_bkg2}
model2.plotOn(frame, Components=p_ph_sig2_set, LineColor="kAzure")
model2.plotOn(frame, Components=p_ph_bkg2_set, LineColor="r")
model2.paramOn(frame, Parameters={Asig2, Abkg2})

sigData.plotOn(frame, MarkerColor="b")
frame.Draw()

pt.Clear()
pt_text3 = [
    "When signal and background",
    "template share one gamma para-",
    "meter per bin, they adapt less.",
    "The errors of the A parameters",
    "also shrink slightly.",
]
for text in pt_text3:
    pt.AddText(text)
pt.DrawClone()


print("Asig [normal ] = {} +/- {}".format(Asig0.getVal(), Asig0.getError()))
print("Asig [BB     ] = {} +/- {}".format(Asig1.getVal(), Asig1.getError()))
print("Asig [BBlight] = {} +/- {}".format(Asig2.getVal(), Asig2.getError()))

can.SaveAs("rf709_BarlowBeeston.png")



# Author: Stefan Wunsch CERN, Vincenzo Eduardo Padulano (UniMiB, CERN) 07/2019

################################################################################
# Copyright (C) 1995-2018, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

import numpy


class ndarray(numpy.ndarray):
    """
    A wrapper class that inherits from `numpy.ndarray` and allows to attach the
    result pointer of the `Take` action in an `RDataFrame` event loop to the
    collection of values returned by that action. See
    https://docs.scipy.org/doc/numpy/user/basics.subclassing.html for more
    information on subclassing numpy arrays.
    """
    def __new__(cls, numpy_array, result_ptr):
        """
        Dunder method invoked at the creation of an instance of this class. It
        creates a numpy array with an `RResultPtr` as an additional
        attribute.
        """
        obj = numpy.asarray(numpy_array).view(cls)
        obj.result_ptr = result_ptr
        return obj

    def __array_finalize__(self, obj):
        """
        Dunder method that fills in the instance default `result_ptr` value.
        """
        if obj is None: return
        self.result_ptr = getattr(obj, "result_ptr", None)



## \file
## \ingroup tutorial_dataframe
## \notebook
## Use the Display action to inspect entry values.
##
## This tutorial shows how to use the Display action
##
## \macro_code
## \macro_output
##
## \date August 2018
## \authors Enrico Guiraud, Danilo Piparo, Enric Tejedor Saavedra (CERN), Massimo Tumolo (Politecnico di Torino)

import ROOT

# Setting up a Dataframe with some data
ROOT.gInterpreter.ProcessLine('''
   unsigned long long y = 1;
   int x = 1;
   double w = 1;
   double z = 1;
   ROOT::RDataFrame df(10);
   auto d = df.Define("y", [&y]() { return y *= 100; }) // A column with ulongs
              .Define("x",
                      [&x]() {
                         return std::vector<int>({x++, x++, x++, x++});
                      })                                // A column with four-elements collection
              .Define("w", [&w]() { return w *= 1.8; }) // A column with doubles
              .Define("z", [&z]() {
                 z *= 1.1;
                 return std::vector<std::vector<double>>({{z, ++z}, {z, ++z}, {z, ++z}});
              }); // A column of matrices
''')

d = ROOT.d

# Preparing the RResultPtr<RDisplay> object with all columns and default number of entries
d1 = d.Display("")
# Preparing the RResultPtr<RDisplay> object with two columns and default number of entries
cols = ROOT.vector('string')(); cols.push_back("x"); cols.push_back("y");
d2 = d.Display(cols)

# Printing the short representations, the event loop will run
print("The following is the representation of all columns with the default nr of entries")
d1.Print()
print("\n\nThe following is the representation of two columns with the default nr of entries")
d2.Print()



# Authors:
# * Lorenzo Moneta 04/2022
# * Harshal Shende 04/2022

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################


r"""
/**
\class Factory
\brief \parblock \endparblock
\htmlonly
<div class="pyrootbox">
\endhtmlonly
## PyROOT

\htmlonly
</div>
\endhtmlonly
*/
"""


from ._utils import _kwargs_to_tmva_cmdargs, cpp_signature


class Factory(object):
    r"""Some member functions of factory that take a TMVA CmdArg as argument also support keyword arguments.
    This applies to factory::BookMethod().
    """

    @cpp_signature(
        "Factory::Factory(TString jobName, TFile *theTargetFile, TString theOption)"
        ": Configurable(theOption), fTransformations('I'), fVerbose(kFALSE), fVerboseLevel(kINFO), fCorrelations(kFALSE),"
        "fROC(kTRUE), fSilentFile(theTargetFile == nullptr), fJobName(jobName), fAnalysisType(Types::kClassification),"
        "fModelPersistence(kTRUE),"
        "Factory::Factory(TString jobName, TString theOption)"
    )
    def __init__(self, *args, **kwargs):
        # Redefinition of `Factory` constructor for keyword arguments.
        # The keywords must correspond to the CmdArg of the constructor function.
        args, kwargs = _kwargs_to_tmva_cmdargs(*args, **kwargs)
        self._init(*args, **kwargs)

    @cpp_signature(
        "Factory::BookMethod( DataLoader *loader, TString theMethodName, TString methodTitle, TString theOption = "
        " );"
    )
    def BookMethod(self, *args, **kwargs):

        r"""factory::BookMethod() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        args, kwargs = _kwargs_to_tmva_cmdargs(*args, **kwargs)
        return self._BookMethod(*args, **kwargs)



def MakeKerasBinary(layer):
    from ROOT.TMVA.Experimental import SOFIE

    input = layer["layerInput"]
    output = layer["layerOutput"]
    fLayerType = layer["layerType"]
    fLayerDType = layer["layerDType"]
    fX1 = input[0]
    fX2 = input[1]
    fY = output[0]
    op = None
    if SOFIE.ConvertStringToType(fLayerDType) == SOFIE.ETensorType.FLOAT:
        if fLayerType == "Add":
            op = SOFIE.ROperator_BasicBinary(float, SOFIE.EBasicBinaryOperator.Add)(fX1, fX2, fY)
        elif fLayerType == "Subtract":
            op = SOFIE.ROperator_BasicBinary(float, SOFIE.EBasicBinaryOperator.Sub)(fX1, fX2, fY)
        else:
            op = SOFIE.ROperator_BasicBinary(float, SOFIE.EBasicBinaryOperator.Mul)(fX1, fX2, fY)
    else:
        raise RuntimeError(
            "TMVA::SOFIE - Unsupported - Operator BasicBinary does not yet support input type " + fLayerDType
        )
    return op



# Author: Enric Tejedor CERN  02/2019

################################################################################
# Copyright (C) 1995-2019, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

from . import pythonization


def _TClonesArray__setitem__(self, key, val):
    raise RuntimeError(
        "You can't set items of a TClonesArray! You can however TClonesArray::ConstructedAt() to access an element that is guaranteed to be constructed, and then modify it in-place."
    )


@pythonization("TClonesArray")
def pythonize_tclonesarray(klass):
    # Parameters:
    # klass: class to be pythonized

    # Add item setter method
    klass.__setitem__ = _TClonesArray__setitem__



# Authors:
# * Jonas Rembser 09/2023
################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################


class RooPlot(object):
    def addObject(self, *args, **kwargs):
        from ROOT._pythonization._memory_utils import declare_cpp_owned_arg

        # Python should transfer the ownership to the RooPlot
        declare_cpp_owned_arg(0, "obj", args, kwargs)

        return self._addObject(*args, **kwargs)



# Author: Stefan Wunsch CERN  09/2019

################################################################################
# Copyright (C) 1995-2019, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

from .. import pythonization


def Compute(self, x):
    import numpy as np

    import ROOT

    # numpy.array is a factory and the actual type of a numpy array is numpy.ndarray
    if isinstance(x, np.ndarray):
        if len(x.shape) == 1:
            x_ = ROOT.VecOps.AsRVec(x)
            y = self._OriginalCompute(x_)
            return np.asarray(y)
        elif len(x.shape) == 2:
            x_ = ROOT.TMVA.Experimental.AsRTensor(x)
            y = self._OriginalCompute(x_)
            return np.asarray(y)
        else:
            raise Exception("Call to Compute can process only numpy arrays of rank 1 or 2.")

    # As fall-through we go to the original compute function and use the error-handling from cppyy
    return self._OriginalCompute(x)


@pythonization("RBDT", ns="TMVA::Experimental", is_prefix=True)
def pythonize_rbdt(klass):
    # Parameters:
    # klass: class to be pythonized

    klass._OriginalCompute = klass.Compute
    klass.Compute = Compute



# Author: Enric Tejedor CERN  04/2019

################################################################################
# Copyright (C) 1995-2019, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

from . import pythonization


def _Draw(self, *args):

    import ROOT

    # Parameters:
    # self: Object being drawn
    # args: arguments for Draw

    self._OriginalDraw(*args)

    # When drawing a TPad, it gets added to the list of primititves of its
    # mother TPad (fMother) with kCanDelete == 1. This means that, when
    # fMother is destructed, it will attempt to destroy its child TPad too.
    # To prevent a double delete, here we instruct the Python proxy of the
    # child C++ TPad being drawn not to destroy the latter (ROOT-10060).
    # 
    # A similar principle is applied to TButton, TColorWheel, TPolyLine3D,
    # TPolyMarker and TPolyMarker3D, whose kCanDelete bit is set in one of
    # their constructors. Later, when being drawn, they are appended to
    # the list of primitives of gPad.
    if self.TestBit(ROOT.kCanDelete):
        ROOT.SetOwnership(self, False)

    self.Draw = self._OriginalDraw

def _init(self, *args):

    import ROOT

    # Parameters:
    # self: Object being initialized
    # args: arguments for __init__

    self._original__init__(*args)

    # TSlider is a special case, since it is appended to gPad already
    # in one of its constructors, after setting kCanDelete.
    # Therefore, we need to set the ownership here and not in Draw
    # (TSlider does not need to be drawn). This is ROOT-10095.
    if self.TestBit(ROOT.kCanDelete):
        ROOT.SetOwnership(self, False)
        # We have already set the ownership while initializing,
        # so we do not need the custom Draw inherited from TPad to
        # do it again in case it is executed.
        self.Draw = self._OriginalDraw

@pythonization([ 'TPad', 'TButton', 'TColorWheel',
                 'TPolyLine3D', 'TPolyMarker', 'TPolyMarker3D' ])
def pythonize_drawables(klass):
    # Parameters:
    # klass: class to be pythonized

    klass._OriginalDraw = klass.Draw
    klass.Draw = _Draw

@pythonization('TSlider')
def pythonize_tslider(klass):
    # Parameters:
    # klass: class to be pythonized

    klass._original__init__ = klass.__init__
    klass.__init__ = _init



## \file
## \ingroup tutorial_fit
## \notebook
## Tutorial for normalized sum of two functions
## Here: a background exponential and a crystalball function
## Parameters can be set:
##  1.   with the TF1 object before adding the function (for 3) and 4))
##  2.  with the TF1NormSum object (first two are the coefficients, then the non constant parameters)
##  3. with the TF1 object after adding the function
##
## Sum can be constructed by:
##  1. by a string containing the names of the functions and/or the coefficient in front
##  2. by a string containg formulas like expo, gaus...
##  3. by the list of functions and coefficients (which are 1 by default)
##  4. by a std::vector for functions and coefficients
##
## \macro_image
## \macro_output
## \macro_code
##
## \author Jonas Rembser, Lorenzo Moneta (C++ version)

import ROOT

nsig = 50000
nbkg = 1000000
nEvents = nsig + nbkg
nBins = 1000

signal_mean = 3.0
f_cb = ROOT.TF1("MyCrystalBall", "crystalball", -5.0, 5.0)
f_exp = ROOT.TF1("MyExponential", "expo", -5.0, 5.0)

# I.:
f_exp.SetParameters(1.0, -0.3)
f_cb.SetParameters(1, signal_mean, 0.3, 2, 1.5)

# CONSTRUCTION OF THE TF1NORMSUM OBJECT ........................................
# 1) :
fnorm_exp_cb = ROOT.TF1NormSum(f_cb, f_exp, nsig, nbkg)
# 4) :

f_sum = ROOT.TF1("fsum", fnorm_exp_cb, -5.0, 5.0, fnorm_exp_cb.GetNpar())

# III.:
parameter_values = fnorm_exp_cb.GetParameters()
f_sum.SetParameters(parameter_values.data())
# Note: in the C++ tutorial, the parameter value sync is done in one line with:
# ```C++
#     f_sum->SetParameters(fnorm_exp_cb->GetParameters().data());
# ```
# However, TF1NormSum::GetParameters() returns an std::vector by value, which
# doesn't survive long enough in Python. That's why we have to explicitly
# assign it to a variable first and can't use a temporary.

f_sum.SetParName(1, "NBackground")
f_sum.SetParName(0, "NSignal")
for i in range(2, f_sum.GetNpar()):
    f_sum.SetParName(i, fnorm_exp_cb.GetParName(i))

# GENERATE HISTOGRAM TO FIT ..............................................................
w = ROOT.TStopwatch()
w.Start()
h_sum = ROOT.TH1D("h_ExpCB", "Exponential Bkg + CrystalBall function", nBins, -5.0, 5.0)
h_sum.FillRandom(f_sum, nEvents)
print("Time to generate {0} events:  ".format(nEvents))
w.Print()

# need to scale histogram with width since we are fitting a density
h_sum.Sumw2()
h_sum.Scale(1.0, "width")

# fit - use Minuit2 if available
ROOT.Math.MinimizerOptions.SetDefaultMinimizer("Minuit2")
c1 = ROOT.TCanvas("Fit", "Fit", 800, 1000)
# do a least-square fit of the spectrum
result = h_sum.Fit(f_sum, "SQ")
result.Print()
h_sum.Draw()
print("Time to fit using ROOT TF1Normsum: ")
w.Print()

# test if parameters are fine
for i, pref in enumerate([nsig, nbkg, signal_mean]):
    if not ROOT.TMath.AreEqualAbs(pref, f_sum.GetParameter(i), f_sum.GetParError(i) * 10.0):
        ROOT.Error(
            "testFitNormSum",
            "Difference found in fitted {0} - difference is {1:.2f} sigma".format(
                f_sum.GetParName(i), (f_sum.GetParameter(i) - pref) / f_sum.GetParError(i)
            ),
        )

ROOT.gStyle.SetOptStat(0)
# add parameters
t1 = ROOT.TLatex(-2.5, 300000, "NSignal = {0:g} #pm {1:g}".format(f_sum.GetParameter(0), f_sum.GetParError(0)))
t2 = ROOT.TLatex(-2.5, 270000, "Nbackgr = {0:g} #pm {1:g}".format(f_sum.GetParameter(1), f_sum.GetParError(1)))
t1.Draw()
t2.Draw()

c1.SaveAs("fitNormSum.png")



# Author: Enric Tejedor CERN  02/2019

################################################################################
# Copyright (C) 1995-2019, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

# Searching


def _contains(self, o):
    # Relies on TObject::FindObject
    # Parameters:
    # - self: object where to search
    # - o: object to be searched in self
    # Returns:
    # - True if self contains o
    return bool(self.FindObject(o))


# Comparison operators


def _lt(self, o):
    import ROOT

    if isinstance(o, ROOT.TObject):
        return self.Compare(o) == -1
    else:
        return NotImplemented


def _le(self, o):
    import ROOT

    if isinstance(o, ROOT.TObject):
        return self.Compare(o) <= 0
    else:
        return NotImplemented


def _gt(self, o):
    import ROOT

    if isinstance(o, ROOT.TObject):
        return self.Compare(o) == 1
    else:
        return NotImplemented


def _ge(self, o):
    import ROOT

    if isinstance(o, ROOT.TObject):
        return self.Compare(o) >= 0
    else:
        return NotImplemented


def pythonize_tobject():
    import ROOT

    klass = ROOT.TObject

    # Allow 'obj in container' syntax for searching
    klass.__contains__ = _contains

    # Inject comparison operators
    klass.__lt__ = _lt
    klass.__le__ = _le
    klass.__gt__ = _gt
    klass.__ge__ = _ge


# Instant pythonization (executed at `import ROOT` time), no need of a
# decorator. This is a core class that is instantiated before cppyy's
# pythonization machinery is in place.
pythonize_tobject()



# Author: Pawan Johnson, Vincenzo Eduardo Padulano, Enric Tejedor  CERN  07/2022

################################################################################
# Copyright (C) 1995-2022, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################
import re
import typing

from .._numbadeclare import _NumbaDeclareDecorator


class FunctionJitter:
    """
    This class allows to jit a python callable with Numba, being able to infer the signature of the function from the types of the RDF columns.
    It takes a python callable, an RDF and optionally a column list and dictionary containing any other arguments for the callable.
    It determines the function signature and Declares the function and returns the corresponding function call.

    Attributes
    ----------
    rdf: RDataFrame object on which the function's column arguments are in
    col_names: The list of columns in rdf
    func: The python callable that is to be jitted.
    return_type: Return type of above callable
    params: List of parameters of the above callable
    func_args: Dict to store the value of each param
    args_info: Information about the type and value of each param
    func_sign: Contains the function signature
    func_call: Contains the function call as a string.

    Example:
    rdf = ROOT.RDataFrame(5).Define("x", "(double) rdfentry_")
    def x_greater_than_2(x):
        return x>2
    fj = FunctionJitter(rdf)
    func_call = fj.jit(x_greater_than_2)
    fil1 = rdf.Filter( "Numba::" + func_call, "x is greater than 2")

    """

    # Variable to store previous functions so as to not rejit.
    function_cache = {}
    lambda_function_counter = 0  # Counter to name the lambda functions

    def __init__(self, rdf: "RDataFrame") -> None:  # noqa F821
        self.rdf = rdf
        self.col_names: typing.List[str] = rdf.GetColumnNames()
        self.func: typing.Callable
        self.return_type: str
        self.params: typing.List[str]
        self.func_args: typing.Dict(str, any)
        self.args_info: typing.Dict(str, (str, any))
        self.func_sign: str
        self.func_call: str

    def find_type(self, x):
        """
        Function to determine the type of a variable x.
        If x is a string. It maps it to the corresponding column's type in RDF. (String constants are not supported)
        Else it determines the fundamental type of x. (int, float, bool)
        Else it is an a numpy array and maps it to corresponding RVec.
        Otherwise flags a type error
        Args:
            `   x: Variable whose type is to be determined

        Returns:
            type of the variable x
        """
        try:
            import numpy as np
        except ImportError:
            raise ImportError("Failed to import numpy during call to determine function signature.")
        from ._rdf_conversion_maps import FUNDAMENTAL_PYTHON_TYPES, NUMPY_TO_TREE, TREE_TO_NUMBA

        if isinstance(x, str):
            # Can be string constant or can be column name
            if x in self.col_names:  # If x is a column
                t = self.rdf.GetColumnType(x)
                if t in TREE_TO_NUMBA:  # The column is a fundamental type from tree
                    return TREE_TO_NUMBA[t]

                match = re.match(r"([\w:]+)<(.+)>", t)
                if match:
                    container_type, inner_type = match.groups()
                    container_type = container_type.strip()
                    inner_type = inner_type.strip()
                    inner_mapped = TREE_TO_NUMBA.get(inner_type, inner_type)
                    return f"{container_type}<{inner_mapped}>"
                else:
                    return t
            else:
                return "str"
                #! Numba Declare does not support "string" type. Check _numbadeclare.Thus, Cannot pass string constants to the filter/Defines..
        elif type(x) in FUNDAMENTAL_PYTHON_TYPES:
            return FUNDAMENTAL_PYTHON_TYPES[type(x)]
        elif isinstance(x, np.ndarray):
            if x.dtype.type in NUMPY_TO_TREE:
                return "RVec<" + NUMPY_TO_TREE[x.dtype.type] + ">"
            else:
                raise TypeError(f"Support for {x.dtype.type} arrays is not yet supported.")
        #! Need to work out how to map things like tuples, dicts, lists...
        else:
            raise TypeError(f"Type of {type(x).__name__}:  {x} cannot be jitted.")

    def find_function_params(self, func):
        """
        Function to create a list of parameters func needs.
        Updates the class to have a new attribute params which is a list of the parameters of func.

        Arguments:
            func: A python callable

        """
        import inspect

        func_sign = inspect.signature(func)
        # Find the Return type
        if func_sign.return_annotation is inspect.Signature.empty:
            self.return_type = None
        else:
            self.return_type = str(func_sign.return_annotation)
        self.func = func
        # List of input parameters for function
        # ALl the input parameters the function needs.
        self.params = list(func_sign.parameters.keys())
        self.args_info = {}

    def generate_func_args(self, cols_list, extra_args):
        """
        Function to create a dictionary mapping the parameters of the function to the corresponding columns or values.
        Updates the class to have a new attribute func_args which contains the above mapping.

        Arguments:
            cols_list: A python list that contains the columns which the function uses.
            extra_args: A dictionary containing any extra parameters the function requires.
        """
        self.func_args = {}
        n_params = len(self.params)
        n_cols = len(cols_list)
        n_constants = len(extra_args)

        if n_cols > 0:
            # Check to see if all the parameters have been provided.
            if n_params != n_cols + n_constants:
                raise ValueError(
                    "Not Enough values provided in the column list and extra_args. The function required {} parameters only {} provided.".format(
                        n_params, n_cols + n_constants
                    )
                )

        # Mapping the column list to the first input parameters of the function.
        for idx, p in enumerate(self.params):
            if idx < n_cols:
                self.func_args[p] = cols_list[idx]
        # Extra args supersedes col_list
        self.func_args = {**self.func_args, **extra_args}

    def find_function_signature(self):
        """
        Calculates the function signature.
        Updates the class with a new attribute args_info which contains the mapping of the parameter of the function to its type and value.
        args_info:  dict{parameter_name(str): (type, value)}
        """
        func_args = self.func_args
        for p in self.params:
            if p in func_args:  # the parameter value has been given in func_args
                value_of_p = func_args[p]
                type_of_p = self.find_type(value_of_p)
                # Bool(s) in python are represented as True/False but in C++ are true/false. The following if statements are to account for that
                if type(value_of_p) is bool:
                    if value_of_p:
                        value_of_p = "true"
                    else:
                        value_of_p = "false"
            else:  # the parameter was not in func_args. Thus this parameter has to be mapped to a column of rdf
                if p not in self.col_names:
                    raise Exception(
                        f"Unable to map function argument {p} to a column.\nUse correct name of column or pass a list of column names."
                    )
                value_of_p = p
                type_of_p = self.find_type(p)
            self.args_info[p] = (type_of_p, value_of_p)

    def generate_function_call(self):
        """
        Generates a function call and signature. Gives a unique name to the function if it is a lambda function.
        Updates the class with new attributes func_call and func_sign to hold them,
        """
        func = self.func
        if func.__name__ == "<lambda>":
            func.__name__ = f"_lambda_func_number_{FunctionJitter.lambda_function_counter}"
            FunctionJitter.lambda_function_counter += 1
        self.func_call = f"{func.__name__}({', '.join(str(arg_info[1]) for arg_info in self.args_info.values())})"
        self.func_sign = [str(self.args_info[p][0]) for p in self.params]

    def get_function_params_args_call(self, func, cols_list, extra_args):
        """
        Function to generate the function params, args, signature and call.
        """
        self.find_function_params(func)
        self.generate_func_args(cols_list, extra_args)
        self.find_function_signature()
        self.generate_function_call()

    def jit_function(self, func, cols_list, extra_args):
        """
        Jits the provided function using ROOT's NumbaDeclare.
        Also checks if the function was jitted earlier in which case it won't jit again but if signature does not match. It raises an error.

        Arguments:
        func: A python callable
        cols_list: A list of columns of RDF on which func depends on.
        extra_args: A dict of extra arguments that func requires.
        """

        if func.__name__ in FunctionJitter.function_cache:
            func_call, func_sign = FunctionJitter.function_cache[func.__name__]
            self.get_function_params_args_call(func, cols_list, extra_args)
            if self.func_sign != func_sign:
                raise ValueError("Trying to re-use a function. Do not change function signature.")
            return self.func_call

        self.get_function_params_args_call(func, cols_list, extra_args)
        _NumbaDeclareDecorator(self.func_sign, self.return_type)(self.func)
        FunctionJitter.function_cache[self.func.__name__] = (self.func_call, self.func_sign)
        return self.func_call


def remove_fn_name_from_signature(signature):
    """
    Removes the function name from a signature string.
    The signature is expected to be in the form of "return_type function_name(type param1, type param2, ...)".
    """
    if "(" not in signature or ")" not in signature:
        raise ValueError(f"Invalid signature format: {signature}")

    return signature[: signature.index(" ")] + signature[signature.index("(") :]


def get_cpp_overload_from_templ_proxy(func, types=None):
    """
    Gets the C++ overload from a cppyy TemplateProxy using the input
    and template arguments types.
    """
    signature = ", ".join(types) if types else ""
    template_args = func.__template_args__[1:-1] if func.__template_args__ else ""
    return func.__overload__(signature, template_args)


def get_column_types(rdf, cols):
    return [rdf.GetColumnType(col) for col in cols]


def get_overload_based_on_args(overload_types, types):
    """
    Gets the C++ overloads for a function based on the provided signatures and types.
    """
    if not isinstance(overload_types, dict):
        raise TypeError("Overload types must be a dictionary.")

    if not isinstance(types, list) or not all(isinstance(t, str) for t in types):
        raise TypeError("Types must be a list of strings.")

    if len(overload_types) == 1:
        # If there is only one signature, return it directly
        return next(iter(overload_types))

    for full_sig, info in overload_types.items():
        input_types = info.get("input_types", ())

        if len(input_types) != len(types):
            continue

        match = True
        for expected, actual in zip(input_types, types):
            # Loose match: require actual to appear somewhere in expected (e.g. "int" in "const int&")
            if actual not in expected:
                match = False
                break

        if match:
            return full_sig

    raise ValueError(
        f"No matching overload found for types: {types} in signatures: {overload_types.keys()}. Please check the function overloads."
    )


def _get_cpp_signature(func, rdf, cols):
    """
    Gets the C++ signature of a cppyy callable.
    """
    import ROOT

    if isinstance(func, ROOT._cppyy.types.Template):
        func = get_cpp_overload_from_templ_proxy(func, get_column_types(rdf, cols))

    if not isinstance(func, ROOT._cppyy.types.Function):
        raise TypeError(f"Expected a cppyy callable, got {type(func).__name__}")

    overload_types = func.func_overloads_types
    matched_overload = get_overload_based_on_args(overload_types, get_column_types(rdf, cols))
    return remove_fn_name_from_signature(matched_overload)


def _to_std_function(func, rdf, cols):
    """
    Converts a cppyy callable to std::function.
    """
    import ROOT

    if not isinstance(func, ROOT._cppyy.types.Function) and not isinstance(func, ROOT._cppyy.types.Template):
        raise TypeError(f"Expected a cppyy callable, got {type(func).__name__}")

    signature = _get_cpp_signature(func, rdf, cols)
    return ROOT.std.function(signature)


def _handle_cpp_callables(func, original_template, *args, rdf=None, cols=None):
    """
    Checks whether the callable `func` is a cppyy proxy of one of these:
    1. C++ functor
    2. std::function
    3. C++ free function

    Cases 1 and 2 above are supported by cppyy, so we can just invoke the original
    cppyy TemplateProxy (Filter or Define) with the callable as argument.
    For case 3, we need to convert the callable to a std::function
    before invoking the original cppyy TemplateProxy.

    Prior to the invocation of the original cppyy TemplateProxy, though, we
    need to explicitly instantiate the template using the type of the `func`
    callable. Implicit instantiation won't work since the original
    implementation of Filter and Define is replaced by a pythonization.

    Arguments:
        func: callable to be checked
        original_rdf_method: cppyy proxy for Filter or Define
        args: arguments passed by the user to Filter or Define
    Returns:
        RDataFrame: RDataFrame node if `func` can be classified in one of the
    three cases above, None otherwise
    """
    import ROOT

    is_cpp_functor = lambda: isinstance(getattr(func, "__call__", None), ROOT._cppyy.types.Function)  # noqa E731

    is_std_function = lambda: isinstance(getattr(func, "target_type", None), ROOT._cppyy.types.Function)  # noqa E731

    # handle free functions
    if callable(func) and not is_cpp_functor() and not is_std_function():
        try:
            func = _to_std_function(func, rdf, cols)
        except TypeError as e:
            if "Expected a cppyy callable" in str(e):
                pass  # this function is not convertible to std::function, move on to the next check
            else:
                raise

    if is_cpp_functor() or is_std_function():
        return original_template[type(func)](*args)


def _PyFilter(rdf, callable_or_str, *args, extra_args={}):
    """
    Filters the entries of RDF according to a given condition.
    Arguments:
        callable_or_str: The condition according to which the RDF is to be filtered.
            It can be either a python callable or a c style string.
        *args:
            Can be at most two of them.
            a. List of columns that the callable will receive as argument.
                If not provided then it tries maps the name of the parameter to a column name of the RDF.
            b. The name of the Filter as a string.
        extra_args: non-columnar arguments to be passed to the callable.
    Returns:
        RDataFrame: result of transforming the rdf argument by applying the requested filter operation.

    Examples:
    1. rdf.Filter(lambda x,y: x>y)
        Filter using a python callable.
    2. rdf.Filter(lambda x,y: x>y, ["y", "x"])
        Filter using a python callable with column list passed explicitly.
        When the column list is passed, the function maps the starting input parameters to those columns instead.
        Here the function param x is mapped to column y, and the second function param y is mapped to column x.
        This is equivalent to the filter "y>x".
    3. rdf.Filter(lambda col1, const1: col1>const1, ["x"], extra_args = {"const1":0.5})
        Filter using a python callable with column list passed explicitly and extra_args.
        Here the function parameter col1 is mapped to col "x" of the RDF, and the const1 parameter is mapped to the float value 0.5
        This is equivalent to the filter x>0.5"
    4. y = 0.5
       def x_more_than_y(x):
            x>y
       rdf.Filter(x_more_than_y)
       Here the value of y is captured from scope. Thus this is equivalent to the filter "x>0.5".
       Note: Any modifications to the value of y will not be reflected in the function as y would be treated as a compile time constant during the first jit.


    """
    if isinstance(callable_or_str, str):  # If string argument is passed. Invoke the Original Filters.
        return rdf._OriginalFilter(callable_or_str, *args)
    # The 1st argument is either a string or a python callable.
    if not callable(callable_or_str):
        raise TypeError(
            f"The first argument of a Filter operation should be a callable. {type(callable_or_str).__name__} object is not callable."
        )

    if len(args) > 2:
        raise TypeError(f"Filter takes at most 3 positional arguments but {len(args) + 1} were given")

    func = callable_or_str
    col_list = []
    filter_name = ""

    if len(args) == 1:
        if isinstance(args[0], list):
            col_list = args[0]
        elif isinstance(args[0], str):
            filter_name = args[0]
        else:
            raise ValueError(f"Argument should be either 'list' or 'str', not {type(args[0]).__name__}.")

    elif len(args) == 2:
        if isinstance(args[0], list) and isinstance(args[1], str):
            col_list = args[0]
            filter_name = args[1]
        else:
            raise ValueError(
                f"Arguments should be ('list', 'str',) not ({type(args[0]).__name__, type(args[1]).__name__}."
            )

    rdf_node = _handle_cpp_callables(func, rdf._OriginalFilter, func, *args, rdf=rdf, cols=col_list)
    if rdf_node is not None:
        return rdf_node

    jitter = FunctionJitter(rdf)
    func.__annotations__["return"] = (
        "bool"  # return type for Filters is bool # Note: You can keep double and Filter still works.
    )

    func_call = jitter.jit_function(func, col_list, extra_args)
    return rdf._OriginalFilter("Numba::" + func_call, filter_name)


def _PyDefine(rdf, col_name, callable_or_str, cols=[], extra_args={}):
    """
    Defines a new column in the RDataFrame.
    Arguments:
    1. col_name: The name of the new column to be defined
    2. callable_or_str: The definition of the contents of the new columns
        It can be either a python callable or a c style string.
    3. cols: list of columns that the callable will receive as argument.
                If not provided then it tries maps the name of the parameter to a column name of the RDF.
    4. extra_args: non-columnar arguments to be passed to the callable.
    Returns:
        RDataFrame: rdf with new column defined

    Examples:
    1. rdf.Define("x", lambda: np.random.rand())
        Define using a python lambda
    2. rdf.Define("x2", lambda x: x*x, ["y"])
        Maps the function argument to column y.
    3. scale = 0.5
        def x_scaled(x):
            return x*scale
       rdf.Define("X_scaled", x_scaled)
       The scale is captured from scope and treated as a compile time constant during first jit.
       Note: Any changes to scale will not be reflected in the jitted function as jitting is a one time operation.
    4. def x_scaled(x, u):
            return x*u
       rdf.Define("x_scaled", x_scaled, extra_args = {"u":0.5})

    """
    if not isinstance(col_name, str):
        raise TypeError(
            f"First argument of Define must be a valid string for the new column name. {type(col_name).__name__} is not a string."
        )

    if isinstance(callable_or_str, str):  # If string argument is passed. Invoke the Original Define.
        return rdf._OriginalDefine(col_name, callable_or_str)

    if not callable(callable_or_str):  # The 2st argument is either a string or a python callable.
        raise TypeError(
            f"The second argument of a Define operation should be a callable. {type(callable_or_str).__name__} object is not callable."
        )

    if not isinstance(cols, list):
        raise TypeError(f"Define takes a column list as third arguments but {type(cols).__name__} was given.")

    func = callable_or_str
    rdf_node = _handle_cpp_callables(func, rdf._OriginalDefine, col_name, func, cols, rdf=rdf, cols=cols)
    if rdf_node is not None:
        return rdf_node

    jitter = FunctionJitter(rdf)
    func_call = jitter.jit_function(func, cols, extra_args)
    return rdf._OriginalDefine(col_name, "Numba::" + func_call)



def MakeKerasELU(layer):
    """
    Create a Keras-compatible exponential linear Unit (ELU) activation operation using SOFIE framework.

    This function takes a dictionary representing a layer and its attributes and
    constructs a Keras-compatible ELU activation operation using the SOFIE framework.
    ELU is an activation function that modifies only the negative part of ReLU by
    applying an exponential curve. It allows small negative values instead of zeros.

    Parameters:
    layer (dict): A dictionary containing layer information including input, output,
                  and data type, which must be float.

    Returns:
    ROperator_Elu: A SOFIE framework operator representing the ELU activation operation.
    """
    from ROOT.TMVA.Experimental import SOFIE

    finput = layer["layerInput"]
    foutput = layer["layerOutput"]
    fLayerDType = layer["layerDType"]
    fLayerInputName = finput[0]
    fLayerOutputName = foutput[0]
    attributes = layer["layerAttributes"]
    if "alpha" in attributes.keys():
        fAlpha = attributes["alpha"]
    else:
        fAlpha = 1.0
    if SOFIE.ConvertStringToType(fLayerDType) == SOFIE.ETensorType.FLOAT:
        op = SOFIE.ROperator_Elu("float")(fAlpha, fLayerInputName, fLayerOutputName)
        return op
    else:
        raise RuntimeError("TMVA::SOFIE - Unsupported - Operator Relu does not yet support input type " + fLayerDType)



# Authors:
# * Jonas Rembser 06/2021
# * Harshal Shende 06/2021

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################


from ._utils import _kwargs_to_roocmdargs, cpp_signature


class RooAbsRealLValue(object):
    r"""Some member functions of RooAbsRealLValue that take a RooCmdArg as argument also support keyword arguments.
    So far, this applies to RooAbsRealLValue::createHistogram and RooAbsRealLValue::frame.
    For example, the following code is equivalent in PyROOT:
    \code{.py}
    # Directly passing a RooCmdArg:
    frame = x.frame(ROOT.RooFit.Name("xframe"), ROOT.RooFit.Title("RooPlot with decorations"), ROOT.RooFit.Bins(40))

    # With keyword arguments:
    frame = x.frame(Name="xframe", Title="RooPlot with decorations", Bins=40)
    \endcode
    """

    __cpp_name__ = 'RooAbsRealLValue'

    @cpp_signature(
        "TH1 *RooAbsRealLValue::createHistogram(const char *name,"
        "    const RooCmdArg& arg1={}, const RooCmdArg& arg2={},"
        "    const RooCmdArg& arg3={}, const RooCmdArg& arg4={},"
        "    const RooCmdArg& arg5={}, const RooCmdArg& arg6={},"
        "    const RooCmdArg& arg7={}, const RooCmdArg& arg8={}) const ;"
    )
    def createHistogram(self, *args, **kwargs):
        r"""The RooAbsRealLValue::createHistogram() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        import ROOT

        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        out = self._createHistogram(*args, **kwargs)
        ROOT.SetOwnership(out, True)
        return out

    @cpp_signature(
        "RooPlot *RooAbsRealLValue::frame(const RooCmdArg& arg1, const RooCmdArg& arg2={},"
        "    const RooCmdArg& arg3={}, const RooCmdArg& arg4={}, const RooCmdArg& arg5={},"
        "    const RooCmdArg& arg6={}, const RooCmdArg& arg7={}, const RooCmdArg& arg8={}) const ;"
    )
    def frame(self, *args, **kwargs):
        r"""The RooAbsRealLValue::frame() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        # Redefinition of `RooAbsRealLValue.frame` for keyword arguments.
        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        frame = self._frame(*args, **kwargs)

        # Add a Python reference to the plot variable to the RooPlot because
        # the plot variable needs to survive at least as long as the plot.
        frame._plotVar = self

        return frame



## \file
## \ingroup tutorial_roofit_main
## \notebook
## 'BASIC FUNCTIONALITY' RooFit tutorial macro #105
## Demonstration of binding ROOT Math functions as RooFit functions
## and pdfs
##
## \macro_image
## \macro_code
## \macro_output
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C version)

import ROOT

# Bind ROOT TMath::Erf C function
# ---------------------------------------------------

# Bind one-dimensional ROOT.TMath.Erf function as ROOT.RooAbsReal function
x = ROOT.RooRealVar("x", "x", -3, 3)
erf = ROOT.RooFit.bindFunction("erf", ROOT.TMath.Erf, x)

# Print erf definition
erf.Print()

# Plot erf on frame
frame1 = x.frame(Title="TMath.Erf bound as ROOT.RooFit function")
erf.plotOn(frame1)

# Bind ROOT::Math::beta_pdf C function
# -----------------------------------------------------------------------

# Bind pdf ROOT.Math.Beta with three variables as ROOT.RooAbsPdf function
x2 = ROOT.RooRealVar("x2", "x2", 0, 0.999)
a = ROOT.RooRealVar("a", "a", 5, 0, 10)
b = ROOT.RooRealVar("b", "b", 2, 0, 10)
beta = ROOT.RooFit.bindPdf("beta", ROOT.Math.beta_pdf, x2, a, b)

# Perf beta definition
beta.Print()

# Generate some events and fit
data = beta.generate({x2}, 10000)
beta.fitTo(data, PrintLevel=-1)

# Plot data and pdf on frame
frame2 = x2.frame(Title="ROOT.Math.Beta bound as ROOT.RooFit pdf")
data.plotOn(frame2)
beta.plotOn(frame2)

# Bind ROOT TF1 as RooFit function
# ---------------------------------------------------------------

# Create a ROOT TF1 function
fa1 = ROOT.TF1("fa1", "sin(x)/x", 0, 10)

# Create an observable
x3 = ROOT.RooRealVar("x3", "x3", 0.01, 20)

# Create binding of TF1 object to above observable
rfa1 = ROOT.RooFit.bindFunction(fa1, x3)

# Print rfa1 definition
rfa1.Print()

# Make plot frame in observable, TF1 binding function
frame3 = x3.frame(Title="TF1 bound as ROOT.RooFit function")
rfa1.plotOn(frame3)

c = ROOT.TCanvas("rf105_funcbinding", "rf105_funcbinding", 1200, 400)
c.Divide(3)
c.cd(1)
ROOT.gPad.SetLeftMargin(0.15)
frame1.GetYaxis().SetTitleOffset(1.6)
frame1.Draw()
c.cd(2)
ROOT.gPad.SetLeftMargin(0.15)
frame2.GetYaxis().SetTitleOffset(1.6)
frame2.Draw()
c.cd(3)
ROOT.gPad.SetLeftMargin(0.15)
frame3.GetYaxis().SetTitleOffset(1.6)
frame3.Draw()

c.SaveAs("rf105_funcbinding.png")



## \file
## \ingroup tutorial_hist
## \notebook
## Example that shows custom dashed lines on the lower plot, specified by a vector of floats.
##
## By default, dashed lines are drawn at certain points. You can either disable them, or specify
## where you want them to appear.
## Inspired by the tutorial of Paul Gessinger.
##
## \macro_image
## \macro_code
##
## \author Alberto Ferro

import ROOT

ROOT.gStyle.SetOptStat(0)

c1 = ROOT.TCanvas("c1", "fit residual simple")
h1 = ROOT.TH1D("h1", "h1", 50, -5, 5)
h1.FillRandom("gaus", 2000)
h1.Fit("gaus")
h1.GetXaxis().SetTitle("x")
h1.GetYaxis().SetTitle("y")

rp1 = ROOT.TRatioPlot(h1)

lines = ROOT.std.vector('double')()
for i in range(-3,4):lines.push_back(i)
rp1.SetGridlines(lines)

rp1.Draw()
rp1.GetLowerRefGraph().SetMinimum(-4)
rp1.GetLowerRefGraph().SetMaximum(4)

c1.Update()



# \file
# \ingroup tutorial_roostats
# \notebook
# StandardFrequentistDiscovery
#
#  This is a standard demo that can be used with any ROOT file
#  prepared in the standard way.  You specify:
#  - name for input ROOT file
#  - name of workspace inside ROOT file that holds model and data
#  - name of ModelConfig that specifies details for calculator tools
#  - name of dataset
#
#  With default parameters the macro will attempt to run the
#  standard hist2workspace example and read the ROOT file
#  that it produces.
#
# \macro_image
# \macro_output
# \macro_code
#
# \authors Sven Kreiss, Kyle Cranmer (C++ version), and P. P. (Python translation)


import ROOT


def StandardFrequentistDiscovery(
    infile="",
    workspaceName="channel1",
    modelConfigNameSB="ModelConfig",
    dataName="obsData",
    toys=1000,
    poiValueForBackground=0.0,
    poiValueForSignal=1.0,
):

    # The workspace contains the model for s+b. The b model is "autogenerated"
    # by copying s+b and setting the one parameter of interest to zero.
    # To keep the script simple, multiple parameters of interest or different
    # functional forms of the b model are not supported.

    # for now, assume there is only one parameter of interest, and these are
    # its values:

    # -------------------------------------------------------
    # First part is just to access a user-defined file
    # or create the standard example file if it doesn't exist
    filename = ""
    if infile == "":
        filename = "results/example_channel1_GammaExample_model.root"
        fileExist = not ROOT.gSystem.AccessPathName(filename)  # note opposite return code
        # if file does not exists generate with histfactory
        if not fileExist:
            # Normally this would be run on the command line
            print(f"will run standard hist2workspace example")
            ROOT.gROOT.ProcessLine(".!  prepareHistFactory .")
            ROOT.gROOT.ProcessLine(".!  hist2workspace config/example.xml")
            print(f"\n\n---------------------")
            print(f"Done creating example input")
            print(f"---------------------\n\n")

    else:
        filename = infile

    # Try to open the file
    file = ROOT.TFile.Open(filename)

    # if input file was specified but not found, quit
    if not file:
        print(f"StandardRooStatsDemoMacro: Input file {filename} is not found")
        return -1

    # -------------------------------------------------------
    # Tutorial starts here
    # -------------------------------------------------------

    mn_t = ROOT.TStopwatch()
    mn_t.Start()

    # get the workspace out of the file
    w = file.Get(workspaceName)
    if not w:
        print(f"workspace not found")
        return -1.0

    # get the modelConfig out of the file
    mc = w[modelConfigNameSB]

    # get the data out of the file
    data = w[dataName]

    # make sure ingredients are found
    if not data or not mc:
        w.Print()
        print(f"data or ModelConfig was not found")
        return -1.0

    firstPOI = mc.GetParametersOfInterest().first()
    firstPOI.setVal(poiValueForSignal)
    mc.SetSnapshot(mc.GetParametersOfInterest())
    # create null model
    mcNull = mc.Clone("ModelConfigNull")
    firstPOI.setVal(poiValueForBackground)
    mcNull.SetSnapshot(mcNull.GetParametersOfInterest().snapshot())

    # ----------------------------------------------------
    # Configure a ProfileLikelihoodTestStat and a SimpleLikelihoodRatioTestStat
    # to use simultaneously with ToyMCSampler
    plts = ROOT.RooStats.ProfileLikelihoodTestStat(mc.GetPdf())
    plts.SetOneSidedDiscovery(True)
    plts.SetVarName("q_0/2")

    # ----------------------------------------------------
    # configure the ToyMCImportanceSampler with two test statistics
    toymcs = ROOT.RooStats.ToyMCSampler(plts, 50)

    # Since this tool needs to throw toy MC the PDF needs to be
    # extended or the tool needs to know how many entries in a dataset
    # per pseudo experiment.
    # In the 'number counting form' where the entries in the dataset
    # are counts, and not values of discriminating variables, the
    # datasets typically only have one entry and the PDF is not
    # extended.
    if not mc.GetPdf().canBeExtended():
        if data.numEntries() == 1:
            toymcs.SetNEventsPerToy(1)
        else:
            print(f"Not sure what to do about this model")

    # instantiate the calculator
    freqCalc = ROOT.RooStats.FrequentistCalculator(data, mc, mcNull, toymcs)
    freqCalc.SetToys(toys, toys)  # null toys, alt toys

    # Run the calculator and print result
    freqCalcResult = freqCalc.GetHypoTest()
    freqCalcResult.GetNullDistribution().SetTitle("b only")
    freqCalcResult.GetAltDistribution().SetTitle("s+b")
    freqCalcResult.Print()
    pvalue = freqCalcResult.NullPValue()

    # stop timing
    mn_t.Stop()
    print(f"total CPU time: ", mn_t.CpuTime())
    print(f"total real time: ", mn_t.RealTime())

    # plot
    c1 = ROOT.TCanvas()
    plot = ROOT.RooStats.HypoTestPlot(freqCalcResult, 100, -0.49, 9.51)
    plot.SetLogYaxis(True)
    plot.Draw()
    c1.Update()
    c1.Draw()
    c1.SaveAs("StandardFrequentistDiscovery.1.png")

    # adding chi2 to plot a different plot.
    # plot 1 and plot 2 have problems at HypoTestPlot.AddTF1
    c2 = ROOT.TCanvas("myc2", "myc2")
    nPOI = 1
    # g =  TF1("g", TString.Format("1*ROOT::Math::chisquared_pdf(2*x,{},0)".format( nPOI)), 0, 9) # doesn´t work properly
    scale_ctte = 0.2
    ROOT.gInterpreter.Declare(
        f"""TF1 *g = new TF1("g", TString::Format("{scale_ctte}*ROOT::Math::chisquared_pdf(2*x,%d,0)", {nPOI}), 0, 9);"""
    )
    g = ROOT.g
    g.SetLineColor(ROOT.kBlack)
    g.SetLineStyle(7)
    plot2 = ROOT.RooStats.HypoTestPlot(freqCalcResult, 100, -0.49, 9.51)
    tmptitle = f"#chi^{{2}}(2x,{nPOI})"
    plot2.AddTF1(g, tmptitle, "SAME")
    plot2.SetLogYaxis(True)
    plot2.Draw()

    c2.Update()
    c2.Draw()
    c2.SaveAs("StandardFrequentistDiscovery.2.png")

    return pvalue


StandardFrequentistDiscovery(
    infile="",
    workspaceName="channel1",
    modelConfigNameSB="ModelConfig",
    dataName="obsData",
    toys=1000,
    poiValueForBackground=0.0,
    poiValueForSignal=1.0,
)



# Authors:
# * Jonas Rembser 06/2021
# * Harshal Shende 06/2021

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################


from ._utils import _dict_to_flat_map, _kwargs_to_roocmdargs, cpp_signature


class RooDataHist(object):
    r"""Constructor of RooDataHist takes a RooCmdArg as argument also supports keyword arguments.
    For example, the following code is equivalent in PyROOT:
    \code{.py}
    # Directly passing a RooCmdArg:
    dh = ROOT.RooDataHist("dh", "dh", ROOT.RooArgList(x), ROOT.RooFit.Import("SampleA", histo))

    # With keyword arguments:
    dh = ROOT.RooDataHist("dh", "dh", ROOT.RooArgList(x), Import=("SampleA", histo))
    \endcode
    """

    __cpp_name__ = 'RooDataHist'

    @cpp_signature(
        [
            "RooDataHist(std::string_view name, std::string_view title, const RooArgList& vars, RooCategory& indexCat, std::map<std::string,TH1*> histMap, Double_t initWgt=1.0) ;",
            "RooDataHist(std::string_view name, std::string_view title, const RooArgList& vars, RooCategory& indexCat, std::map<std::string,RooDataHist*> dhistMap, Double_t wgt=1.0) ;",
        ]
    )
    def __init__(self, *args, **kwargs):
        r"""The RooDataHist constructor is pythonized with the command argument pythonization and for converting python dict to std::map.
        The keywords must correspond to the CmdArg of the constructor function.
        The instances in dict must correspond to the template argument in std::map of the constructor.
        """
        # Redefinition of `RooDataHist` constructor for keyword arguments and converting python dict to std::map.
        if len(args) > 4 and isinstance(args[4], dict):
            args = list(args)
            args[4] = _dict_to_flat_map(args[4], {"std::string": ["RooDataHist*", "TH1*"]})

        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        self._init(*args, **kwargs)

    @property
    def shape(self):
        shape = []

        for i_var, var in enumerate(self.get()):
            shape.append(var.numBins())

        def get_len_from_shape():
            n_size_check = 1
            for n in shape:
                n_size_check = n_size_check * n
            return n_size_check

        # Cross check that the product of the shape values is equal to the full
        # length.
        assert get_len_from_shape() == len(self)

        return tuple(shape)

    def __len__(self):
        return self.numEntries()

    def _to_array(self, buffer):
        # Helper to create a numpy array from a raw array pointer.
        #
        # Note: The data is copied.
        #
        # Args:
        #     buffer (cppyy.LowLevelView):
        #         The pointer to the beginning of the array data, usually
        #         obtained from a C++ function that returns a `double *`.
        #
        # Returns:
        #     numpy.ndarray
        import numpy as np

        # Check if buffer is a nullptr by using implicit conversion from
        # `nullptr` to Bool (for some reason comparing with ROOT.nullptr
        # doesn't work).
        if not buffer:
            return None
        a = np.copy(np.frombuffer(buffer, dtype=np.float64, count=len(self)))
        return a.reshape(self.shape)

    def _var_is_category(self):
        # Returns a list with a Bool for each dimension of the histogram that
        # flags whether the variable in this dimension is a RooAbsCategory.
        import ROOT

        out = []

        for i_var, var in enumerate(self.get()):
            if isinstance(var, ROOT.RooAbsReal):
                out.append(False)
            elif isinstance(var, ROOT.RooAbsCategory):
                out.append(True)
            else:
                raise TypeError("Variables in RooDataHist should either be RooAbsReal or RooAbsCategory.")

        return out

    def _weight_error_low(self):
        # Returns the low weight errors as numpy arrays.
        return self._to_array(self.wgtErrLoArray())

    def _weight_error_high(self):
        # Returns the high weight errors as numpy arrays.
        return self._to_array(self.wgtErrHiArray())

    def _weight_error(self):
        # Returns the low and high weight errors as numpy arrays.
        return self.weight_error_low(), self.weight_error_high()

    def _weights_squared_sum(self):
        # Returns the sum of squared weights that were used to fill each bin.
        return self._to_array(self.sumW2Array())

    @staticmethod
    def from_numpy(hist_weights, variables, bins=None, ranges=None, weights_squared_sum=None, name=None, title=None):
        r"""Create a RooDataHist from numpy arrays.

        Note: The argument structure was inspired by numpy.histogramdd.

        Args:
            hist_weights (numpy.ndarray): The multidimensional histogram bin
                                          weights.
            bins (list): The bin specification, where each element is either:
                           * a numpy array describing the monotonically
                             increasing bin edges along each dimension.
                           * a scalar value for the number of bins (in this
                             case, the corresponding item in the `ranges`
                             argument must be filled)
                           * `None` for a category dimension or if you want to
                              use the default binning of the RooFit variable
            variables (RooArgSet, or list/tuple of RooAbsArgs):
                Specification of the variables in the RooDataHist, will be
                forwarded to the RooDataHist constructor. Both real values and
                categories are supported.
            ranges (list): An optional list specifying the variable range
                           limits. Each element is either:
                             * `None` if a full bin edges array is given to
                               `bins` or for a category dimension
                             * a tuple with two values corresponding to the
                               minimum and maximum values
            weights_squared_sum (numpy.ndarray):
                The sum of squared weights of the original samples that were
                used to fill the histogram. If the input weights are from a
                weighted histogram, this parameter is no longer optional.
            name (str): Name of the RooDataSet, `None` is equivalent to an
                        empty string.
            title (str): Title of the RooDataSet, `None` is equivalent to an
                         empty string.

        Returns:
            RooDataHist
        """
        import numpy as np

        import ROOT

        name = "" if name is None else name
        title = "" if title is None else title

        n_dim = len(variables)

        if bins is None:
            bins = [None] * n_dim

        if ranges is None:
            ranges = [None] * n_dim

        # name for internal binning that is created for the RooDataHist to adapt
        binning_name = "_roodataset_from_numpy_binning"

        def set_binning(var, bins, ranges):
            if bins is None:
                binning = var.getBinning().clone()
            elif np.isscalar(bins):
                if ranges is None:
                    raise ValueError("If a scalar number of bins is given, you must also provide an explicit range.")
                binning = ROOT.RooUniformBinning(ranges[0], ranges[1], bins)
            else:
                binning = ROOT.RooBinning(len(bins) - 1, bins)

            var.setBinning(binning, binning_name)

        for i_var, var in enumerate(variables):
            if isinstance(var, ROOT.RooAbsReal):
                set_binning(var, bins[i_var], ranges[i_var])

        datahist = ROOT.RooDataHist(name, title, variables, binning_name)

        if len(datahist) != len(hist_weights):
            raise ValueError("Length of hist_weights array doesn't match the size of the RooDataHist.")

        if weights_squared_sum is None:
            if not np.allclose(hist_weights, hist_weights.round()):
                raise ValueError(
                    "Your input histogram has non-integer weights! You must also provide weights_squared_sum to provide the complete information to RooDataHist.from_numpy()."
                )
        else:
            if len(datahist) != len(weights_squared_sum):
                raise ValueError("Length of weights_squared_sum array doesn't match the size of the RooDataHist.")

        for i_bin in range(len(datahist)):
            if weights_squared_sum is not None:
                # some reverse-computation that can't be avoided with the current C++ RooDataHist interface
                wgt_err = np.sqrt(weights_squared_sum[i_bin])
            else:
                wgt_err = -1

            datahist.set(i_bin, hist_weights[i_bin], wgt_err)

        return datahist

    def to_numpy(self):
        r"""Converts the weights and bin edges of a RooDataHist to numpy arrays.

        Note: The output structure was inspired by numpy.histogramdd.

        Returns:
            weight (numpy.ndarray): The weights for each histrogram bin.
            bin_edges (list): A list of `n_dim` arrays describing the bin edges
                              for each dimension. For dimensions of category
                              types, the list element is `None`.
        """
        import numpy as np

        bin_edges = []

        var_is_category = self._var_is_category()

        for i_var, var in enumerate(self.get()):

            if var_is_category[i_var]:
                bin_edges.append(None)
            else:
                binning = self.getBinnings()[i_var]
                bin_edges.append(
                    np.copy(np.frombuffer(binning.array(), dtype=np.float64, count=binning.numBoundaries()))
                )

        return self._to_array(self.weightArray()), bin_edges



## \file
## \ingroup tutorial_roostats
## \notebook
## Limits: number counting experiment with uncertainty on both the background rate and signal efficiency.
##
## The usage of a Confidence Interval Calculator to set a limit on the signal is illustrated
##
## \macro_image
## \macro_output
## \macro_code
##
## \date June 2022
## \authors Artem Busorgin, Kyle Cranmer (C++ version)

import ROOT

# --------------------------------------
# An example of setting a limit in a number counting experiment with uncertainty on background and signal

# to time the macro
t = ROOT.TStopwatch()
t.Start()

# --------------------------------------
# The Model building stage
# --------------------------------------
wspace = ROOT.RooWorkspace()
wspace.factory(
    "Poisson::countingModel(obs[150,0,300], " "sum(s[50,0,120]*ratioSigEff[1.,0,3.],b[100]*ratioBkgEff[1.,0.,3.]))"
)  # counting model
wspace.factory("Gaussian::sigConstraint(gSigEff[1,0,3],ratioSigEff,0.05)")  # 5% signal efficiency uncertainty
wspace.factory("Gaussian::bkgConstraint(gSigBkg[1,0,3],ratioBkgEff,0.2)")  # 10% background efficiency uncertainty
wspace.factory("PROD::modelWithConstraints(countingModel,sigConstraint,bkgConstraint)")  # product of terms
wspace.Print()

modelWithConstraints = wspace["modelWithConstraints"]  # get the model
obs = wspace["obs"]  # get the observable
s = wspace["s"]  # get the signal we care about
b = wspace["b"]  # get the background and set it to a constant.  Uncertainty included in ratioBkgEff
b.setConstant()

ratioSigEff = wspace["ratioSigEff"]  # get uncertain parameter to constrain
ratioBkgEff = wspace["ratioBkgEff"]  # get uncertain parameter to constrain
constrainedParams = {ratioSigEff, ratioBkgEff}  # need to constrain these in the fit (should change default behavior)

gSigEff = wspace["gSigEff"]  # global observables for signal efficiency
gSigBkg = wspace["gSigBkg"]  # global obs for background efficiency
gSigEff.setConstant()
gSigBkg.setConstant()

# Create an example dataset with 160 observed events
obs.setVal(160.0)
dataOrig = ROOT.RooDataSet("exampleData", "exampleData", {obs})
dataOrig.add(obs)

# not necessary
modelWithConstraints.fitTo(dataOrig, Constrain=constrainedParams, PrintLevel=-1)

# Now let's make some confidence intervals for s, our parameter of interest
modelConfig = ROOT.RooStats.ModelConfig(wspace)
modelConfig.SetPdf(modelWithConstraints)
modelConfig.SetParametersOfInterest({s})
modelConfig.SetNuisanceParameters(constrainedParams)
modelConfig.SetObservables(obs)
modelConfig.SetGlobalObservables({gSigEff, gSigBkg})
modelConfig.SetName("ModelConfig")
wspace.Import(modelConfig)
wspace.Import(dataOrig)
wspace.SetName("w")
# wspace.writeToFile("rs101_ws.root")

# Make sure we reference the data in the workspace from now on
data = wspace[dataOrig.GetName()]

# First, let's use a Calculator based on the Profile Likelihood Ratio
plc = ROOT.RooStats.ProfileLikelihoodCalculator(data, modelConfig)
plc.SetTestSize(0.05)
lrinterval = plc.GetInterval()

# Let's make a plot
dataCanvas = ROOT.TCanvas("dataCanvas")
dataCanvas.Divide(2, 1)
dataCanvas.cd(1)

plotInt = ROOT.RooStats.LikelihoodIntervalPlot(lrinterval)
plotInt.SetTitle("Profile Likelihood Ratio and Posterior for S")
plotInt.Draw()

# Second, use a Calculator based on the Feldman Cousins technique
fc = ROOT.RooStats.FeldmanCousins(data, modelConfig)
fc.UseAdaptiveSampling(True)
fc.FluctuateNumDataEntries(False)  # number counting analysis: dataset always has 1 entry with N events observed
fc.SetNBins(100)  # number of points to test per parameter
fc.SetTestSize(0.05)
# fc.SaveBeltToFile(True) # optional
fcint = fc.GetInterval()

fit = modelWithConstraints.fitTo(data, Save=True, PrintLevel=-1)

# Third, use a Calculator based on Markov Chain monte carlo
# Before configuring the calculator, let's make a ProposalFunction
# that will achieve a high acceptance rate
ph = ROOT.RooStats.ProposalHelper()
ph.SetVariables(fit.floatParsFinal())
ph.SetCovMatrix(fit.covarianceMatrix())
ph.SetUpdateProposalParameters(True)
ph.SetCacheSize(100)
pdfProp = ph.GetProposalFunction()

mc = ROOT.RooStats.MCMCCalculator(data, modelConfig)
mc.SetNumIters(20000)  # steps to propose in the chain
mc.SetTestSize(0.05)  # 95% CL
mc.SetNumBurnInSteps(40)  # ignore first N steps in chain as "burn in"
mc.SetProposalFunction(pdfProp)
mc.SetLeftSideTailFraction(0.5)  # find a "central" interval
mcInt = mc.GetInterval()

# Get Lower and Upper limits from Profile Calculator
print("Profile lower limit on s = ", lrinterval.LowerLimit(s))
print("Profile upper limit on s = ", lrinterval.UpperLimit(s))

# Get Lower and Upper limits from FeldmanCousins with profile construction
if fcint:
    fcul = fcint.UpperLimit(s)
    fcll = fcint.LowerLimit(s)
    print("FC lower limit on s = ", fcll)
    print("FC upper limit on s = ", fcul)
    fcllLine = ROOT.TLine(fcll, 0, fcll, 1)
    fculLine = ROOT.TLine(fcul, 0, fcul, 1)
    fcllLine.SetLineColor("kRed")
    fculLine.SetLineColor("kRed")
    fcllLine.Draw("same")
    fculLine.Draw("same")
    dataCanvas.Update()

# Plot MCMC interval and print some statistics
mcPlot = ROOT.RooStats.MCMCIntervalPlot(mcInt)
mcPlot.SetLineColor("kMagenta")
mcPlot.SetLineWidth(2)
mcPlot.Draw("same")

mcul = mcInt.UpperLimit(s)
mcll = mcInt.LowerLimit(s)
print("MCMC lower limit on s = ", mcll)
print("MCMC upper limit on s = ", mcul)
print("MCMC Actual confidence level: ", mcInt.GetActualConfidenceLevel())

# 3-d plot of the parameter points
dataCanvas.cd(2)

# also plot the points in the markov chain
chainData = mcInt.GetChainAsDataSet()

print("plotting the chain data - nentries = ", chainData.numEntries())
chain = ROOT.RooStats.GetAsTTree("chainTreeData", "chainTreeData", chainData)
chain.SetMarkerStyle(6)
chain.SetMarkerColor("kRed")

chain.Draw("s:ratioSigEff:ratioBkgEff", "nll_MarkovChain_local_", "box")  # 3-d box proportional to posterior

# the points used in the profile construction
parScanData = fc.GetPointsToScan()
print("plotting the scanned points used in the frequentist construction - npoints = ", parScanData.numEntries())

gr = ROOT.TGraph2D(parScanData.numEntries())
for ievt in range(parScanData.numEntries()):
    evt = parScanData.get(ievt)
    x = evt.getRealValue("ratioBkgEff")
    y = evt.getRealValue("ratioSigEff")
    z = evt.getRealValue("s")
    gr.SetPoint(ievt, x, y, z)

gr.SetMarkerStyle(24)
gr.Draw("P SAME")

# print timing info
t.Stop()
t.Print()

dataCanvas.SaveAs("rs101_limitexample.png")



## \file
## \ingroup tutorial_dataframe
## \notebook -nodraw
## Create data from scratch with RDataFrame.
##
## This tutorial shows how to create a dataset from scratch with RDataFrame
##
## \macro_code
##
## \date June 2017
## \author Danilo Piparo (CERN)

import ROOT

# We create an empty data frame of 100 entries
df = ROOT.RDataFrame(100)

# We now fill it with random numbers
ROOT.gRandom.SetSeed(1)
df_1 = df.Define("rnd", "gRandom->Gaus()")

# And we write out the dataset on disk
df_1.Snapshot("randomNumbers", "df008_createDataSetFromScratch_py.root")



## \file
## \ingroup tutorial_roofit_main
## \notebook
## 'LIKELIHOOD AND MINIMIZATION' RooFit tutorial macro #605
##
## Working with the profile likelihood estimator
##
## \macro_image
## \macro_code
## \macro_output
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C version)


import ROOT


# Create model and dataset
# -----------------------------------------------

# Observable
x = ROOT.RooRealVar("x", "x", -20, 20)

# Model (intentional strong correlations)
mean = ROOT.RooRealVar("mean", "mean of g1 and g2", 0, -10, 10)
sigma_g1 = ROOT.RooRealVar("sigma_g1", "width of g1", 3)
g1 = ROOT.RooGaussian("g1", "g1", x, mean, sigma_g1)

sigma_g2 = ROOT.RooRealVar("sigma_g2", "width of g2", 4, 3.0, 6.0)
g2 = ROOT.RooGaussian("g2", "g2", x, mean, sigma_g2)

frac = ROOT.RooRealVar("frac", "frac", 0.5, 0.0, 1.0)
model = ROOT.RooAddPdf("model", "model", [g1, g2], [frac])

# Generate 1000 events
data = model.generate({x}, 1000)

# Construct plain likelihood
# ---------------------------------------------------

# Construct unbinned likelihood
nll = model.createNLL(data, NumCPU=2)

# Minimize likelihood w.r.t all parameters before making plots
ROOT.RooMinimizer(nll).migrad()

# Plot likelihood scan frac
frame1 = frac.frame(Bins=10, Range=(0.01, 0.95), Title="LL and profileLL in frac")
nll.plotOn(frame1, ShiftToZero=True)

# Plot likelihood scan in sigma_g2
frame2 = sigma_g2.frame(Bins=10, Range=(3.3, 5.0), Title="LL and profileLL in sigma_g2")
nll.plotOn(frame2, ShiftToZero=True)

# Construct profile likelihood in frac
# -----------------------------------------------------------------------

# The profile likelihood estimator on nll for frac will minimize nll w.r.t
# all floating parameters except frac for each evaluation

pll_frac = nll.createProfile({frac})

# Plot the profile likelihood in frac
pll_frac.plotOn(frame1, LineColor="r")

# Adjust frame maximum for visual clarity
frame1.SetMinimum(0)
frame1.SetMaximum(3)

# Construct profile likelihood in sigma_g2
# -------------------------------------------------------------------------------

# The profile likelihood estimator on nll for sigma_g2 will minimize nll
# w.r.t all floating parameters except sigma_g2 for each evaluation
pll_sigmag2 = nll.createProfile({sigma_g2})

# Plot the profile likelihood in sigma_g2
pll_sigmag2.plotOn(frame2, LineColor="r")

# Adjust frame maximum for visual clarity
frame2.SetMinimum(0)
frame2.SetMaximum(3)

# Make canvas and draw ROOT.RooPlots
c = ROOT.TCanvas("rf605_profilell", "rf605_profilell", 800, 400)
c.Divide(2)
c.cd(1)
ROOT.gPad.SetLeftMargin(0.15)
frame1.GetYaxis().SetTitleOffset(1.4)
frame1.Draw()
c.cd(2)
ROOT.gPad.SetLeftMargin(0.15)
frame2.GetYaxis().SetTitleOffset(1.4)
frame2.Draw()

c.SaveAs("rf605_profilell.png")
